from __future__ import annotations

import builtins
import re
import sys
from io import BufferedReader, IOBase
from struct import Struct, unpack
from types import MethodType
from typing import Any, Callable, ClassVar, Dict, List, Literal, Optional, Tuple, Union

reNot0 = re.compile(b"(.*?)\x00", re.S)

SYS_ENDIAN = "<" if sys.byteorder == "little" else ">"
Endianess = Literal["<", ">"]

# generate unpack and unpack_from functions
TYPE_PARAM_SIZE_LIST = [
    ("short", "h"),
    ("u_short", "H"),
    ("int", "i"),
    ("u_int", "I"),
    ("long", "q"),
    ("u_long", "Q"),
    ("half", "e"),
    ("float", "f"),
    ("double", "d"),
]

MEMORY_FUNCTIONS: Dict[Endianess, Dict[str, Callable[["EndianBinaryReader_Memoryview"], Any]]] = {"<": {}, ">": {}}
STREAM_FUNCTIONS: Dict[Endianess, Dict[str, Callable[["EndianBinaryReader_Streamable"], Any]]] = {"<": {}, ">": {}}


class EndianBinaryReader:
    Length: int
    Position: int
    BaseOffset: int
    _endian: Endianess
    _function_map: ClassVar[Dict[Endianess, Dict[str, Callable]]]

    def __new__(
        cls,
        item: Union[bytes, bytearray, memoryview, IOBase, str],
        endian: Endianess = ">",
        offset: int = 0,
    ):
        if isinstance(item, (bytes, bytearray, memoryview)):
            obj = super(EndianBinaryReader, cls).__new__(EndianBinaryReader_Memoryview)  # type: ignore
        elif isinstance(item, IOBase):
            obj = super(EndianBinaryReader, cls).__new__(EndianBinaryReader_Streamable)  # type: ignore
        elif isinstance(item, str):
            obj = super(EndianBinaryReader, cls).__new__(EndianBinaryReader_Streamable_LocalFile)  # type: ignore
        elif isinstance(item, EndianBinaryReader):
            item = item.stream if isinstance(item, EndianBinaryReader_Streamable) else item.view
            return EndianBinaryReader(item, endian, offset)
        elif hasattr(item, "read"):
            if hasattr(item, "seek") and hasattr(item, "tell"):
                obj = super(EndianBinaryReader, cls).__new__(EndianBinaryReader_Streamable)
            else:
                item = item.read()
                obj = super(EndianBinaryReader, cls).__new__(EndianBinaryReader_Memoryview)
        else:
            raise TypeError("Unsupported type for EndianBinaryReader: %s" % type(item))
        return obj

    def __init__(self, item, endian: Endianess = ">", offset: int = 0):
        self._endian = ""  # type: ignore
        self.endian = endian
        self.BaseOffset = offset
        self.Position = 0

    @property
    def endian(self) -> Endianess:
        return self._endian

    @endian.setter
    def endian(self, value: Endianess):
        if value not in ("<", ">"):
            raise ValueError("Invalid endian")
        if value != self._endian:
            for func_name, func in self._function_map[value].items():
                setattr(self, func_name, MethodType(func, self))
            self._endian = value

    @property
    def bytes(self) -> builtins.bytes:
        # implemented by Streamable and Memoryview versions
        return b""

    def read(self, size: Optional[int] = -1, /) -> builtins.bytes:
        # implemented by Streamable and Memoryview versions
        return b""

    def read_byte(self) -> int:
        return unpack(self.endian + "b", self.read(1))[0]

    def read_u_byte(self) -> int:
        return unpack(self.endian + "B", self.read(1))[0]

    def read_bytes(self, num: int) -> builtins.bytes:
        return self.read(num)

    def read_short(self) -> int:
        return unpack(self.endian + "h", self.read(2))[0]

    def read_int(self) -> int:
        return unpack(self.endian + "i", self.read(4))[0]

    def read_long(self) -> int:
        return unpack(self.endian + "q", self.read(8))[0]

    def read_u_short(self) -> int:
        return unpack(self.endian + "H", self.read(2))[0]

    def read_u_int(self) -> int:
        return unpack(self.endian + "I", self.read(4))[0]

    def read_u_long(self) -> int:
        return unpack(self.endian + "Q", self.read(8))[0]

    def read_float(self) -> float:
        return unpack(self.endian + "f", self.read(4))[0]

    def read_double(self) -> float:
        return unpack(self.endian + "d", self.read(8))[0]

    def read_boolean(self) -> bool:
        return bool(unpack(self.endian + "?", self.read(1))[0])

    def read_string(self, size: Optional[int] = None) -> str:
        if size is None:
            return self.read_string_to_null()
        else:
            raw = self.read_bytes(size)
            return raw.decode("utf8", "surrogateescape")

    def read_string_to_null(self, max_length: int = 32767) -> str:
        ret = []
        c = b""
        while c != b"\0" and len(ret) < max_length and self.Position != self.Length:
            ret.append(c)
            c = self.read(1)
            if not c:
                raise ValueError("Unterminated string: %r" % ret)
        return b"".join(ret).decode("utf8", "surrogateescape")

    def read_aligned_string(self) -> str:
        length = self.read_int()
        if 0 < length <= self.Length - self.Position:
            string_data = bytes(self.read_bytes(length))
            result = string_data.decode("utf8", "surrogateescape")
            self.align_stream()
            return result
        return ""

    def align_stream(self, alignment=4):
        self.Position += (alignment - self.Position % alignment) % alignment

    def read_byte_array(self) -> builtins.bytes:
        return self.read(self.read_int())

    def read_array(self, command: Callable, length: int) -> list:
        return [command() for _ in range(length)]

    def read_array_struct(self, param: str, length: Optional[int] = None) -> tuple:
        if length is None:
            length = self.read_int()
        struct = Struct(f"{self.endian}{length}{param}")
        return struct.unpack(self.read(struct.size))

    def read_boolean_array(self, length: Optional[int] = None) -> Tuple[bool, ...]:
        return self.read_array_struct("?", length)

    def read_u_byte_array(self, length: Optional[int] = None) -> Tuple[int, ...]:
        return self.read_array_struct("B", length)

    def read_u_short_array(self, length: Optional[int] = None) -> Tuple[int, ...]:
        return self.read_array_struct("h", length)

    def read_short_array(self, length: Optional[int] = None) -> Tuple[int, ...]:
        return self.read_array_struct("H", length)

    def read_int_array(self, length: Optional[int] = None) -> Tuple[int, ...]:
        return self.read_array_struct("i", length)

    def read_u_int_array(self, length: Optional[int] = None) -> Tuple[int, ...]:
        return self.read_array_struct("I", length)

    def read_long_array(self, length: Optional[int] = None) -> Tuple[int, ...]:
        return self.read_array_struct("q", length)

    def read_u_long_array(self, length: Optional[int] = None) -> Tuple[int, ...]:
        return self.read_array_struct("Q", length)

    def read_float_array(self, length: Optional[int] = None) -> Tuple[float, ...]:
        return self.read_array_struct("f", length)

    def read_double_array(self, length: Optional[int] = None) -> Tuple[float, ...]:
        return self.read_array_struct("d", length)

    def read_string_array(self) -> List[str]:
        return self.read_array(self.read_aligned_string, self.read_int())

    def real_offset(self) -> int:
        """Returns offset in the underlying file.
        (Not working with unpacked streams.)
        """
        return self.BaseOffset + self.Position

    def read_the_rest(self, obj_start: int, obj_size: int) -> builtins.bytes:
        """Returns the rest of the current reader bytes."""
        return self.read_bytes(obj_size - (self.Position - obj_start))


class EndianBinaryReader_Memoryview(EndianBinaryReader):
    __slots__ = ("view", "_endian", "BaseOffset", "Position", "Length")
    _endian: Endianess
    view: memoryview
    _function_map = MEMORY_FUNCTIONS

    def __init__(self, view, endian: Endianess = ">", offset: int = 0):
        super().__init__(view, endian=endian, offset=offset)
        self.view = memoryview(view)
        self.Length = len(view)

    @property
    def bytes(self):
        return self.view.tobytes()

    def dispose(self) -> None:
        self.view.release()

    def read(self, size: Optional[int] = -1, /):
        if not size:
            return b""
        if size == -1:
            size = self.Length - self.Position
        ret = self.view[self.Position : self.Position + size]
        self.Position += size
        return ret.tobytes()

    def read_array_struct(self, param: str, length: Optional[int] = None) -> tuple:
        if length is None:
            length = self.read_int()
        struct = Struct(f"{self.endian}{length}{param}")
        value = struct.unpack_from(self.view, self.Position)
        self.Position += struct.size
        return value

    def read_aligned_string(self) -> str:
        length = self.read_int()
        if 0 < length <= self.Length - self.Position:
            string_data = self.read_bytes(length)
            result = bytes(string_data).decode("utf8", "surrogateescape")
            self.align_stream()
            return result
        return ""

    def read_string_to_null(self, max_length: int = 32767) -> str:
        match = reNot0.search(self.view, self.Position, self.Position + max_length)
        if not match:
            if self.Position + max_length >= self.Length:
                raise Exception("String not terminated")
            else:
                return bytes(self.read_bytes(max_length)).decode("utf8", "surrogateescape")
        ret = match[1].decode("utf8", "surrogateescape")
        self.Position = match.end()
        return ret


class EndianBinaryReader_Streamable(EndianBinaryReader):
    __slots__ = ("stream", "_endian", "BaseOffset")
    stream: BufferedReader
    _function_map = STREAM_FUNCTIONS

    def __init__(self, stream: BufferedReader, endian: Endianess = ">", offset: int = 0):
        self.stream = stream
        super().__init__(stream, endian=endian, offset=offset)
        self.read = self.stream.read

    @property
    def Position(self) -> int:
        return self.stream.tell() - self.BaseOffset

    @Position.setter
    def Position(self, value: int):
        if value < 0:
            raise ValueError("Position cannot be negative")
        self.stream.seek(value + self.BaseOffset)

    @property
    def Length(self):  # type: ignore
        pos = self.Position
        length = self.stream.seek(0, 2) - self.BaseOffset
        self.Position = pos
        return length

    @property
    def bytes(self):
        last_pos = self.Position
        self.Position = 0
        ret = self.read(self.Length)
        self.Position = last_pos  # type: ignore
        return ret

    def dispose(self):
        self.stream.close()
        pass


class EndianBinaryReader_Streamable_LocalFile(EndianBinaryReader_Streamable):
    def __init__(self, path: str, endian: Endianess = ">", offset: int = 0):
        super().__init__(open(path, "rb"), endian=endian, offset=offset)

    def __del__(self):
        self.stream.close()


for endian_s in ("<", ">"):
    for reader_type_name, struct_type_char in TYPE_PARAM_SIZE_LIST:
        func_name = f"read_{reader_type_name}"
        struct = Struct(f"{endian_s}{struct_type_char}")

        def memory_read_func(self: EndianBinaryReader_Memoryview, /, struct=struct):
            value = struct.unpack_from(self.view, self.Position)[0]
            self.Position += struct.size
            return value

        def stream_read_func(self: EndianBinaryReader_Streamable, /, struct=struct):
            return struct.unpack(self.stream.read(struct.size))[0]

        MEMORY_FUNCTIONS[endian_s][func_name] = memory_read_func
        STREAM_FUNCTIONS[endian_s][func_name] = stream_read_func



from typing import List, Optional, Tuple

from PIL.Image import Image

from UnityPy.classes import PPtr
from UnityPy.classes.generated import (
    GUID,
    MonoBehaviour,
    NamedObject,
    Rectf,
    SpriteAtlas,
    SpriteBone,
    SpriteRenderData,
)
from UnityPy.classes.math import Vector2f, Vector4f

class Sprite(NamedObject):
    m_Extrude: int
    m_Name: str
    m_Offset: Vector2f
    m_PixelsToUnits: float
    m_RD: SpriteRenderData
    m_Rect: Rectf
    m_AtlasTags: Optional[List[str]] = None
    m_Bones: Optional[List[SpriteBone]] = None
    m_Border: Optional[Vector4f] = None
    m_IsPolygon: Optional[bool] = None
    m_PhysicsShape: Optional[List[List[Vector2f]]] = None
    m_Pivot: Optional[Vector2f] = None
    m_RenderDataKey: Optional[Tuple[GUID, int]] = None
    m_ScriptableObjects: Optional[List[PPtr[MonoBehaviour]]] = None
    m_SpriteAtlas: Optional[PPtr[SpriteAtlas]] = None

    @property
    def image(self) -> Image: ...



from __future__ import annotations

from typing import TYPE_CHECKING, Dict, Generator, Iterator, List, Tuple, Union

from attrs import define

if TYPE_CHECKING:
    from ..classes import AssetBundle, AssetInfo, Object, PPtr


@define(slots=True)
class ContainerHelper:
    """Helper class to allow multidict containers
    without breaking compatibility with old versions"""

    container: List[Tuple[str, AssetInfo]]
    container_dict: Dict[str, PPtr[Object]]
    path_dict: Dict[int, str]

    def __init__(self, container: Union[List[Tuple[str, AssetInfo]], AssetBundle]) -> None:
        if not isinstance(container, (list)):
            container = container.m_Container
        self.container = container
        self.container_dict = {key: value.asset for key, value in container}
        self.path_dict = {value.asset.path_id: key for key, value in container}

    def items(self) -> Generator[Tuple[str, PPtr[Object]], None, None]:
        return ((key, value.asset) for key, value in self.container)

    def keys(self) -> list[str]:
        return list({key for key, value in self.container})

    def values(self) -> list[PPtr[Object]]:
        return list({value.asset for key, value in self.container})

    def __getitem__(self, key) -> PPtr[Object]:
        return self.container_dict[key]

    def __setitem__(self, key, value) -> None:
        raise NotImplementedError("Assigning to container is not allowed!")

    def __delitem__(self, key) -> None:
        raise NotImplementedError("Deleting from the container is not allowed!")

    def __iter__(self) -> Iterator[str]:
        return iter(self.keys())

    def __len__(self) -> int:
        return len(self.container)

    def __getattr__(self, name: str) -> PPtr[Object]:
        return self.container_dict[name]

    def __str__(self) -> str:
        return f"{{{', '.join(f'{key}: {value}' for key, value in self.items())}}}"

    def __dict__(self) -> Dict[str, PPtr[Object]]:
        return self.container_dict

    def __contains__(self, key: str) -> bool:
        return key in self.container_dict



from . import (
    AudioClipConverter,
    MeshExporter,
    MeshRendererExporter,
    SpriteHelper,
    Texture2DConverter,
)

__all__ = [
    "AudioClipConverter",
    "MeshExporter",
    "MeshRendererExporter",
    "SpriteHelper",
    "Texture2DConverter",
]



from argparse import ArgumentParser

from UnityPy.cli.update_tpk import update_tpk


def main():
    parser = ArgumentParser(
        prog="UnityPy",
        description="UnityPy cli utility",
    )

    subparsers = parser.add_subparsers(title="utils")

    update_tpk_parser = subparsers.add_parser("update_tpk", help="Updates TPK (typetree dump) file")
    update_tpk_parser.set_defaults(func=update_tpk)

    args = parser.parse_args()

    if not hasattr(args, "func"):
        parser.print_help()
        return

    operands = getattr(args, "operands", [])
    args.func(*operands)


if __name__ == "__main__":
    main()



from typing import List, Optional

from PIL.Image import Image

from UnityPy.classes.generated import GLTextureSettings, StreamingInfo, Texture

class Texture2DArray(Texture):
    image_data: bytes
    m_ColorSpace: int
    m_DataSize: int
    m_Depth: int
    m_Format: int
    m_Height: int
    m_IsReadable: bool
    m_MipCount: int
    m_Name: str
    m_TextureSettings: GLTextureSettings
    m_Width: int
    m_DownscaleFallback: Optional[bool] = None
    m_ForcedFallbackFormat: Optional[int] = None
    m_IgnoreMipmapLimit: Optional[bool] = None
    m_IsAlphaChannelOptional: Optional[bool] = None
    m_MipmapLimitGroupName: Optional[str] = None
    m_MipsStripped: Optional[int] = None
    m_StreamData: Optional[StreamingInfo] = None
    m_UsageMode: Optional[int] = None

    @property
    def images(self) -> List[Image]: ...



"""
Definitions for math related classes.
As most calculations involving them are done in numpy,
we define them here as subtypes of np.ndarray, so that casting won't be necessary.
"""

from attrs import define


@define(slots=True)
class Vector2f:
    x: float = 0
    y: float = 0

    def __repr__(self) -> str:
        return f"Vector2f({self.x}, {self.y})"


@define(slots=True)
class Vector3f:
    x: float = 0
    y: float = 0
    z: float = 0

    def __repr__(self) -> str:
        return f"Vector3f({self.x}, {self.y}, {self.z})"


@define(slots=True)
class Vector4f:
    x: float = 0
    y: float = 0
    z: float = 0
    w: float = 0

    def __repr__(self) -> str:
        return f"Vector4f({self.x}, {self.y}, {self.z}, {self.w})"


float3 = Vector3f
float4 = Vector4f


class Quaternionf(Vector4f):
    # TODO: Implement quaternion operations
    def __repr__(self) -> str:
        return f"Quaternion({self.x}, {self.y}, {self.z}, {self.w})"


@define(slots=True)
class Matrix3x4f:
    e00: float
    e01: float
    e02: float
    e03: float
    e10: float
    e11: float
    e12: float
    e13: float
    e20: float
    e21: float
    e22: float
    e23: float


@define(slots=True)
class Matrix4x4f:
    e00: float
    e01: float
    e02: float
    e03: float
    e10: float
    e11: float
    e12: float
    e13: float
    e20: float
    e21: float
    e22: float
    e23: float
    e30: float
    e31: float
    e32: float
    e33: float


@define(slots=True)
class ColorRGBA:
    r: float = 0
    g: float = 0
    b: float = 0
    a: float = 1

    def __init__(self, r: float = 0, g: float = 0, b: float = 0, a: float = 1, rgba: int = -1):
        if rgba != -1:
            r = ((rgba >> 24) & 0xFF) / 255
            g = ((rgba >> 16) & 0xFF) / 255
            b = ((rgba >> 8) & 0xFF) / 255
            a = (rgba & 0xFF) / 255
        # defined by attrs
        self.__attrs_init__(r, g, b, a)  # type: ignore

    @property
    def rgba(self) -> int:
        return int(self.r * 255) << 24 | int(self.g * 255) << 16 | int(self.b * 255) << 8 | int(self.a * 255)

    @rgba.setter
    def rgba(self, value: int):
        self.r = ((value >> 24) & 0xFF) / 255
        self.g = ((value >> 16) & 0xFF) / 255
        self.b = ((value >> 8) & 0xFF) / 255
        self.a = (value & 0xFF) / 255


__all__ = (
    "Vector2f",
    "Vector3f",
    "Vector4f",
    "Quaternionf",
    "Matrix3x4f",
    "Matrix4x4f",
    "ColorRGBA",
    "float3",
    "float4",
)



import builtins
from io import BytesIO, IOBase
from struct import pack
from typing import Callable, Sequence, TypeVar, Union

T = TypeVar("T")


class EndianBinaryWriter:
    endian: str
    Position: int
    stream: IOBase

    def __init__(self, input_: Union[bytes, bytearray, IOBase] = b"", endian: str = ">"):
        if isinstance(input_, (bytes, bytearray)):
            self.stream = BytesIO(input_)
            self.stream.seek(0, 2)
        elif isinstance(input_, IOBase):
            self.stream = input_
        else:
            raise ValueError("Invalid input type - %s." % type(input_))
        self.endian = endian
        self.Position = self.stream.tell()

    @property
    def bytes(self):
        self.stream.seek(0)
        return self.stream.read()

    @property
    def Length(self) -> int:
        pos = self.stream.tell()
        self.stream.seek(0, 2)
        length = self.stream.tell()
        self.stream.seek(pos)
        return length

    def dispose(self):
        self.stream.close()

    def write(self, *args):
        if self.Position != self.stream.tell():
            self.stream.seek(self.Position)
        ret = self.stream.write(*args)
        self.Position = self.stream.tell()
        return ret

    def write_byte(self, value: int):
        self.write(pack(self.endian + "b", value))

    def write_u_byte(self, value: int):
        self.write(pack(self.endian + "B", value))

    def write_bytes(self, value: builtins.bytes):
        return self.write(value)

    def write_short(self, value: int):
        self.write(pack(self.endian + "h", value))

    def write_int(self, value: int):
        self.write(pack(self.endian + "i", value))

    def write_long(self, value: int):
        self.write(pack(self.endian + "q", value))

    def write_u_short(self, value: int):
        self.write(pack(self.endian + "H", value))

    def write_u_int(self, value: int):
        self.write(pack(self.endian + "I", value))

    def write_u_long(self, value: int):
        self.write(pack(self.endian + "Q", value))

    def write_float(self, value: float):
        self.write(pack(self.endian + "f", value))

    def write_double(self, value: float):
        self.write(pack(self.endian + "d", value))

    def write_boolean(self, value: bool):
        self.write(pack(self.endian + "?", value))

    def write_string_to_null(self, value: str):
        self.write(value.encode("utf8", "surrogateescape"))
        self.write(b"\0")

    def write_aligned_string(self, value: str):
        bstring = value.encode("utf8", "surrogateescape")
        self.write_int(len(bstring))
        self.write(bstring)
        self.align_stream(4)

    def align_stream(self, alignment: int = 4):
        pos = self.stream.tell()
        align = (alignment - pos % alignment) % alignment
        self.write(b"\0" * align)

    def write_array(
        self,
        command: Callable[[T], None],
        value: Sequence[T],
        write_length: bool = True,
    ):
        if write_length:
            self.write_int(len(value))
        for val in value:
            command(val)

    def write_byte_array(self, value: builtins.bytes):
        self.write_int(len(value))
        self.write(value)

    def write_boolean_array(self, value: Sequence[bool]):
        self.write_array(self.write_boolean, value)

    def write_u_short_array(self, value: Sequence[int]):
        self.write_array(self.write_u_short, value)

    def write_int_array(self, value: Sequence[int], write_length: bool = False):
        return self.write_array(self.write_int, value, write_length)

    def write_u_int_array(self, value: Sequence[int], write_length: bool = False):
        return self.write_array(self.write_u_int, value, write_length)

    def write_float_array(self, value: Sequence[float], write_length: bool = False):
        return self.write_array(self.write_float, value, write_length)

    def write_string_array(self, value: Sequence[str]):
        self.write_array(self.write_aligned_string, value)



class TypeTreeError(Exception):
    def __init__(self, message, nodes):
        super().__init__(message)
        self.nodes = nodes


class UnityVersionFallbackError(Exception):
    def __init__(self, message):
        super().__init__(message)


class UnityVersionFallbackWarning(UserWarning):
    def __init__(self, message):
        super().__init__(message)



from typing import List, Tuple, Union

from UnityPy.classes import Component, PPtr
from UnityPy.classes.generated import ComponentPair, EditorExtension

class GameObject(EditorExtension):
    m_Component: Union[List[ComponentPair], List[Tuple[int, PPtr[Component]]]]
    m_IsActive: Union[bool, int]
    m_Layer: int
    m_Name: str
    m_Tag: int

    @property
    def m_Components(self) -> List[PPtr[Component]]: ...
    @property
    def m_Animator(self) -> Union[PPtr[Component], None]: ...
    @property
    def m_Animation(self) -> Union[PPtr[Component], None]: ...
    @property
    def m_Transform(self) -> Union[PPtr[Component], None]: ...
    @property
    def m_SkinnedMeshRenderer(self) -> Union[PPtr[Component], None]: ...
    @property
    def m_MeshRenderer(self) -> Union[PPtr[Component], None]: ...
    @property
    def m_MeshFilter(self) -> Union[PPtr[Component], None]: ...



import os
from io import BytesIO
from urllib.request import urlopen
from zipfile import ZipFile

URL = "https://nightly.link/AssetRipper/Tpk/workflows/type_tree_tpk/master/uncompressed_file.zip"
RESOURCE_PATH = os.path.join(os.path.dirname(__file__), "..", "resources")


def update_tpk():
    print("Updating TPK file...")
    print("\tDownloading...")
    with urlopen(URL) as response:
        zip_data = response.read()
    print("\tExtracting...")
    with ZipFile(BytesIO(zip_data)) as zip_file:
        zip_file.extract("uncompressed.tpk", path=RESOURCE_PATH)
    print("\tDone.")


__all__ = ["update_tpk"]



from enum import IntEnum


class TextureFormat(IntEnum):
    Alpha8 = 1
    ARGB4444 = 2
    RGB24 = 3
    RGBA32 = 4
    ARGB32 = 5
    ARGBFloat = 6
    RGB565 = 7
    BGR24 = 8
    R16 = 9
    DXT1 = 10
    DXT3 = 11
    DXT5 = 12
    RGBA4444 = 13
    BGRA32 = 14
    RHalf = 15
    RGHalf = 16
    RGBAHalf = 17
    RFloat = 18
    RGFloat = 19
    RGBAFloat = 20
    YUY2 = 21
    RGB9e5Float = 22
    RGBFloat = 23
    BC6H = 24
    BC7 = 25
    BC4 = 26
    BC5 = 27
    DXT1Crunched = 28
    DXT5Crunched = 29
    PVRTC_RGB2 = 30
    PVRTC_RGBA2 = 31
    PVRTC_RGB4 = 32
    PVRTC_RGBA4 = 33
    ETC_RGB4 = 34
    ATC_RGB4 = 35
    ATC_RGBA8 = 36
    EAC_R = 41
    EAC_R_SIGNED = 42
    EAC_RG = 43
    EAC_RG_SIGNED = 44
    ETC2_RGB = 45
    ETC2_RGBA1 = 46
    ETC2_RGBA8 = 47
    ASTC_RGB_4x4 = 48
    ASTC_RGB_5x5 = 49
    ASTC_RGB_6x6 = 50
    ASTC_RGB_8x8 = 51
    ASTC_RGB_10x10 = 52
    ASTC_RGB_12x12 = 53
    ASTC_RGBA_4x4 = 54
    ASTC_RGBA_5x5 = 55
    ASTC_RGBA_6x6 = 56
    ASTC_RGBA_8x8 = 57
    ASTC_RGBA_10x10 = 58
    ASTC_RGBA_12x12 = 59
    ETC_RGB4_3DS = 60
    ETC_RGBA8_3DS = 61
    RG16 = 62
    R8 = 63
    ETC_RGB4Crunched = 64
    ETC2_RGBA8Crunched = 65
    ASTC_HDR_4x4 = 66
    ASTC_HDR_5x5 = 67
    ASTC_HDR_6x6 = 68
    ASTC_HDR_8x8 = 69
    ASTC_HDR_10x10 = 70
    ASTC_HDR_12x12 = 71
    RG32 = 72
    RGB48 = 73
    RGBA64 = 74
    R8_SIGNED = 75
    RG16_SIGNED = 76
    RGB24_SIGNED = 77
    RGBA32_SIGNED = 78
    R16_SIGNED = 79
    RG32_SIGNED = 80
    RGB48_SIGNED = 81
    RGBA64_SIGNED = 82



from ..generated import Renderer


def export(self, export_dir: str) -> None:
    from ...export import MeshRendererExporter

    MeshRendererExporter.export_mesh_renderer(self, export_dir)


Renderer.export = export

__all__ = ("Renderer",)



from typing import List, Optional, Tuple, Union

from UnityPy.classes import PPtr
from UnityPy.classes.generated import GUID, NamedObject, SerializedShader, Texture

class Shader(NamedObject):
    m_Name: str
    compressedBlob: Optional[List[int]] = None
    compressedLengths: Optional[Union[List[int], List[List[int]]]] = None
    decompressedLengths: Optional[Union[List[int], List[List[int]]]] = None
    decompressedSize: Optional[int] = None
    m_AssetGUID: Optional[GUID] = None
    m_Dependencies: Optional[List[PPtr[Shader]]] = None
    m_NonModifiableTextures: Optional[List[Tuple[str, PPtr[Texture]]]] = None
    m_ParsedForm: Optional[SerializedShader] = None
    m_PathName: Optional[str] = None
    m_Script: Optional[str] = None
    m_ShaderIsBaked: Optional[bool] = None
    m_SubProgramBlob: Optional[List[int]] = None
    offsets: Optional[Union[List[int], List[List[int]]]] = None
    platforms: Optional[List[int]] = None
    stageCounts: Optional[List[int]] = None

    def export(self) -> str: ...



from typing import BinaryIO, Optional, Union

from PIL import Image

from ..generated import Texture2D


def _Texture2d_get_image(self: Texture2D):
    from ...export import Texture2DConverter

    return Texture2DConverter.get_image_from_texture2d(self)


def _Texture2d_set_image(
    self: Texture2D,
    img: Union["Image.Image", str, BinaryIO],
    target_format: Optional[int] = None,
    mipmap_count: int = 1,
):
    from ...export import Texture2DConverter

    if not target_format:
        target_format = self.m_TextureFormat

    if not isinstance(img, Image.Image):
        img = Image.open(img)

    platform = self.object_reader.platform if self.object_reader is not None else 0
    img_data, tex_format = Texture2DConverter.image_to_texture2d(img, target_format, platform, self.m_PlatformBlob)
    self.m_Width = img.width
    self.m_Height = img.height

    if mipmap_count > 1:
        width = self.m_Width
        height = self.m_Height
        re_img = img
        for i in range(mipmap_count - 1):
            width //= 2
            height //= 2
            if width < 4 or height < 4:
                mipmap_count = i + 1
                break
            re_img = re_img.resize((width, height), Image.Resampling.BICUBIC)
            img_data += Texture2DConverter.image_to_texture2d(re_img, target_format)[0]

    # disable mipmaps as we don't store them ourselves by default
    if self.m_MipMap is not None:
        self.m_MipMap = mipmap_count > 1
    if self.m_MipCount is not None:
        self.m_MipCount = mipmap_count

    self.image_data = img_data
    # width * height * channel count
    self.m_CompleteImageSize = len(img_data)  # img.width * img.height * len(img.getbands())
    self.m_TextureFormat = tex_format

    if self.m_StreamData is not None:
        self.m_StreamData.path = ""
        self.m_StreamData.offset = 0
        self.m_StreamData.size = 0


def _Texture2D_get_image_data(self: Texture2D):
    if self.image_data:
        return self.image_data
    if self.m_StreamData:
        from ...helpers.ResourceReader import get_resource_data

        return get_resource_data(
            self.m_StreamData.path,
            self.object_reader.assets_file,
            self.m_StreamData.offset,
            self.m_StreamData.size,
        )
    raise ValueError("No image data found")


Texture2D.image = property(_Texture2d_get_image, _Texture2d_set_image)
Texture2D.set_image = _Texture2d_set_image
Texture2D.get_image_data = _Texture2D_get_image_data


__all__ = ("Texture2D",)



from __future__ import annotations

import struct
from io import BytesIO
from threading import Lock
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Tuple, Union

import astc_encoder
import texture2ddecoder
from PIL import Image

from ..enums import BuildTarget, TextureFormat
from ..helpers import TextureSwizzler

if TYPE_CHECKING:
    from ..classes import Texture2D


TF = TextureFormat

TEXTURE_FORMAT_BLOCK_SIZE_TABLE: Dict[TF, Optional[Tuple[int, int]]] = {}
for tf in TF:
    if tf.name.startswith("ASTC"):
        split = tf.name.rsplit("_", 1)[1].split("x")
        block_size = (int(split[0]), int(split[1]))
    elif tf.name.startswith(("DXT", "BC", "ETC", "EAC")):
        block_size = (4, 4)
    elif tf.name.startswith("PVRTC"):
        block_size = (8 if tf.name.endswith("2") else 4, 4)
    else:
        block_size = None
    TEXTURE_FORMAT_BLOCK_SIZE_TABLE[tf] = block_size


def get_compressed_image_size(width: int, height: int, texture_format: TextureFormat):
    block_size = TEXTURE_FORMAT_BLOCK_SIZE_TABLE[texture_format]
    if block_size is None:
        return (width, height)
    block_width, block_height = block_size

    def pad(value: int, pad_by: int) -> int:
        to_pad = value % pad_by
        if to_pad:
            value += pad_by - to_pad
        return value

    width = pad(width, block_width)
    height = pad(height, block_height)
    return width, height


def pad_image(img: Image.Image, pad_width: int, pad_height: int) -> Image.Image:
    ori_width, ori_height = img.size
    if pad_width == ori_width and pad_height == ori_height:
        return img

    # Paste the original image at the top-left corner
    pad_img = Image.new(img.mode, (pad_width, pad_height))
    pad_img.paste(img, (0, 0))

    # Fill the right border: duplicate the last column
    if pad_width != ori_width:
        right_strip = img.crop((ori_width - 1, 0, ori_width, ori_height))
        right_strip = right_strip.resize((pad_width - ori_width, ori_height), resample=Image.Resampling.NEAREST)
        pad_img.paste(right_strip, (ori_width, 0))

    # Fill the bottom border: duplicate the last row
    if pad_height != ori_height:
        bottom_strip = img.crop((0, ori_height - 1, ori_width, ori_height))
        bottom_strip = bottom_strip.resize((ori_width, pad_height - ori_height), resample=Image.Resampling.NEAREST)
        pad_img.paste(bottom_strip, (0, ori_height))

    # Fill the bottom-right corner with the bottom-right pixel
    if pad_width != ori_width and pad_height != ori_height:
        corner = img.getpixel((ori_width - 1, ori_height - 1))
        corner_img = Image.new(img.mode, (pad_width - ori_width, pad_height - ori_height), color=corner)
        pad_img.paste(corner_img, (ori_width, ori_height))

    return pad_img


def compress_etcpak(data: bytes, width: int, height: int, target_texture_format: TextureFormat) -> bytes:
    import etcpak

    if target_texture_format in [TF.DXT1, TF.DXT1Crunched]:
        return etcpak.compress_bc1(data, width, height)
    elif target_texture_format in [TF.DXT5, TF.DXT5Crunched]:
        return etcpak.compress_bc3(data, width, height)
    elif target_texture_format == TF.BC4:
        return etcpak.compress_bc4(data, width, height)
    elif target_texture_format == TF.BC5:
        return etcpak.compress_bc5(data, width, height)
    elif target_texture_format == TF.BC7:
        return etcpak.compress_bc7(data, width, height, None)
    elif target_texture_format in [TF.ETC_RGB4, TF.ETC_RGB4Crunched, TF.ETC_RGB4_3DS]:
        return etcpak.compress_etc1_rgb(data, width, height)
    elif target_texture_format == TF.ETC2_RGB:
        return etcpak.compress_etc2_rgb(data, width, height)
    elif target_texture_format in [TF.ETC2_RGBA8, TF.ETC2_RGBA8Crunched, TF.ETC2_RGBA1]:
        return etcpak.compress_etc2_rgba(data, width, height)
    else:
        raise NotImplementedError(f"etcpak has no compress function for {target_texture_format.name}")


def compress_astc(data: bytes, width: int, height: int, target_texture_format: TextureFormat) -> bytes:
    astc_image = astc_encoder.ASTCImage(astc_encoder.ASTCType.U8, width, height, 1, data)
    block_size = TEXTURE_FORMAT_BLOCK_SIZE_TABLE[target_texture_format]
    assert block_size is not None, f"failed to get block size for {target_texture_format.name}"
    swizzle = astc_encoder.ASTCSwizzle.from_str("RGBA")

    context, lock = get_astc_context(block_size)
    with lock:
        enc_img = context.compress(astc_image, swizzle)

    return enc_img


def image_to_texture2d(
    img: Image.Image,
    target_texture_format: Union[TextureFormat, int],
    platform: Union[BuildTarget, int] = 0,
    platform_blob: Optional[List[int]] = None,
    flip: bool = True,
) -> Tuple[bytes, TextureFormat]:
    """Converts a PIL Image to Texture2D bytes."""
    if not isinstance(target_texture_format, TextureFormat):
        target_texture_format = TextureFormat(target_texture_format)

    if flip:
        img = img.transpose(Image.Transpose.FLIP_TOP_BOTTOM)

    # defaults
    compress_func = None
    tex_format = TF.RGBA32
    pil_mode = "RGBA"

    # DXT / BC
    if target_texture_format in [TF.DXT1, TF.DXT1Crunched]:
        tex_format = TF.DXT1
        compress_func = compress_etcpak
    elif target_texture_format in [TF.DXT5, TF.DXT5Crunched]:
        tex_format = TF.DXT5
        compress_func = compress_etcpak
    elif target_texture_format in [TF.BC4, TF.BC5, TF.BC7]:
        tex_format = target_texture_format
        compress_func = compress_etcpak
    # ASTC
    elif target_texture_format.name.startswith("ASTC"):
        if "_HDR_" in target_texture_format.name:
            block_size = TEXTURE_FORMAT_BLOCK_SIZE_TABLE[target_texture_format]
            assert block_size is not None
            if img.mode == "RGB":
                tex_format = getattr(TF, f"ASTC_RGB_{block_size[0]}x{block_size[1]}")
            else:
                tex_format = getattr(TF, f"ASTC_RGBA_{block_size[0]}x{block_size[1]}")
        else:
            tex_format = target_texture_format
        compress_func = compress_astc
    # ETC
    elif target_texture_format in [TF.ETC_RGB4, TF.ETC_RGB4Crunched, TF.ETC_RGB4_3DS]:
        if target_texture_format == TF.ETC_RGB4_3DS:
            tex_format = TF.ETC_RGB4_3DS
        else:
            tex_format = target_texture_format
        compress_func = compress_etcpak
    elif target_texture_format == TF.ETC2_RGB:
        tex_format = TF.ETC2_RGB
        compress_func = compress_etcpak
    elif target_texture_format in [TF.ETC2_RGBA8, TF.ETC2_RGBA8Crunched, TF.ETC2_RGBA1]:
        tex_format = TF.ETC2_RGBA8
        compress_func = compress_etcpak
    # A
    elif target_texture_format == TF.Alpha8:
        tex_format = TF.Alpha8
        pil_mode = "A"
    # R - should probably be merged into #A, as pure R is used as Alpha
    # but need test data for this first
    elif target_texture_format in [
        TF.R8,
        TF.R16,
        TF.RHalf,
        TF.RFloat,
        TF.EAC_R,
        TF.EAC_R_SIGNED,
    ]:
        tex_format = TF.R8
        pil_mode = "R"
    # RGBA
    elif target_texture_format in [
        TF.RGB565,
        TF.RGB24,
        TF.BGR24,
        TF.RGB9e5Float,
        TF.PVRTC_RGB2,
        TF.PVRTC_RGB4,
        TF.ATC_RGB4,
    ]:
        tex_format = TF.RGB24
        pil_mode = "RGB"
    # everything else defaulted to RGBA

    width, height = img.width, img.height
    switch_info = None

    if TextureSwizzler.is_switch_swizzled(platform, platform_blob):
        s_tex_format = tex_format
        if tex_format == TF.RGB24:
            s_tex_format = TF.RGBA32
            pil_mode = "RGBA"
        elif tex_format == TF.BGR24:
            s_tex_format = TF.BGRA32
            pil_mode = "BGRA"

        assert platform_blob is not None
        gobs_per_block = TextureSwizzler.get_switch_gobs_per_block(platform_blob)
        block_size = TextureSwizzler.TEXTURE_FORMAT_BLOCK_SIZE_MAP[s_tex_format]
        width, height = TextureSwizzler.get_padded_texture_size(img.width, img.height, *block_size, gobs_per_block)
        switch_info = (block_size, gobs_per_block)

    if compress_func:
        width, height = get_compressed_image_size(width, height, tex_format)
        img = pad_image(img, width, height)
        enc_img = compress_func(img.tobytes("raw", "RGBA"), width, height, tex_format)
    else:
        if switch_info:
            img = pad_image(img, width, height)

        enc_img = img.tobytes("raw", pil_mode)

    if switch_info:
        block_size, gobs_per_block = switch_info
        enc_img = bytes(TextureSwizzler.swizzle(enc_img, width, height, *block_size, gobs_per_block))

    return enc_img, tex_format


def get_image_from_texture2d(
    texture_2d: Texture2D,
    flip: bool = True,
) -> Image.Image:
    """Converts the given Texture2D object to PIL Image."""
    return parse_image_data(
        texture_2d.get_image_data(),
        texture_2d.m_Width,
        texture_2d.m_Height,
        texture_2d.m_TextureFormat,
        getattr(texture_2d.object_reader, "version", (0, 0, 0, 0)),
        getattr(texture_2d.object_reader, "platform", BuildTarget.UnknownPlatform),
        getattr(texture_2d, "m_PlatformBlob", None),
        flip,
    )


def parse_image_data(
    image_data: Union[bytes, bytearray, memoryview],
    width: int,
    height: int,
    texture_format: Union[TextureFormat, int],
    version: Tuple[int, int, int, int],
    platform: Union[BuildTarget, int],
    platform_blob: Optional[List[int]] = None,
    flip: bool = True,
) -> Image.Image:
    """Converts the given image data bytes to PIL Image."""
    if not width or not height:
        return Image.new("RGBA", (0, 0))

    if not image_data:
        raise ValueError("Texture2D has no image data")

    if not isinstance(texture_format, TextureFormat):
        texture_format = TextureFormat(texture_format)

    if platform == BuildTarget.XBOX360 and texture_format in XBOX_SWAP_FORMATS:
        image_data = swap_bytes_for_xbox(image_data)

    original_width, original_height = (width, height)

    if TextureSwizzler.is_switch_swizzled(platform, platform_blob):
        if texture_format == TF.RGB24:
            texture_format = TF.RGBA32
        elif texture_format == TF.BGR24:
            texture_format = TF.BGRA32

        assert platform_blob is not None
        gobs_per_block = TextureSwizzler.get_switch_gobs_per_block(platform_blob)
        block_size = TextureSwizzler.TEXTURE_FORMAT_BLOCK_SIZE_MAP[texture_format]
        width, height = TextureSwizzler.get_padded_texture_size(width, height, *block_size, gobs_per_block)
        image_data = TextureSwizzler.deswizzle(image_data, width, height, *block_size, gobs_per_block)
    else:
        width, height = get_compressed_image_size(width, height, texture_format)

    if not isinstance(image_data, bytes):
        # bytes(bytes item) would cause an unnecessary copy
        image_data = bytes(image_data)

    if "Crunched" in texture_format.name:
        if (
            version[0] > 2017
            or (version[0] == 2017 and version[1] >= 3)  # 2017.3 and up
            or texture_format == TF.ETC_RGB4Crunched
            or texture_format == TF.ETC2_RGBA8Crunched
        ):
            image_data = texture2ddecoder.unpack_unity_crunch(image_data)
        else:
            image_data = texture2ddecoder.unpack_crunch(image_data)

    if texture_format not in CONV_TABLE:
        raise NotImplementedError(f"Not implemented texture format: {texture_format.name}")
    conv_func, conv_args = CONV_TABLE[texture_format]

    img = conv_func(image_data, width, height, *conv_args)

    if original_width != width or original_height != height:
        img = img.crop((0, 0, original_width, original_height))

    if img and flip:
        return img.transpose(Image.Transpose.FLIP_TOP_BOTTOM)

    return img


def swap_bytes_for_xbox(image_data: Union[bytes, bytearray, memoryview]) -> bytearray:
    """Swaps the texture bytes for textures deployed on XBOX360."""
    image_data = bytearray(image_data)
    for i in range(0, len(image_data), 2):
        image_data[i : i + 2] = image_data[i : i + 2][::-1]
    return image_data


def pillow(
    image_data: bytes,
    width: int,
    height: int,
    mode: str,
    codec: str,
    args,
    swap: Optional[Tuple[int, ...]] = None,
) -> Image.Image:
    img = Image.frombytes(mode, (width, height), image_data, codec, args) if width else Image.new(mode, (width, height))
    if swap:
        channels = img.split()
        img = Image.merge(mode, [channels[x] for x in swap])
    return img


def atc(image_data: bytes, width: int, height: int, alpha: bool) -> Image.Image:
    if alpha:
        image_data = texture2ddecoder.decode_atc_rgba8(image_data, width, height)
    else:
        image_data = texture2ddecoder.decode_atc_rgb4(image_data, width, height)

    return Image.frombytes("RGBA", (width, height), image_data, "raw", "BGRA")


def astc(image_data: bytes, width: int, height: int, block_size: tuple) -> Image.Image:
    image = astc_encoder.ASTCImage(astc_encoder.ASTCType.U8, width, height, 1)
    texture_size = calculate_astc_compressed_size(width, height, block_size)
    if len(image_data) < texture_size:
        raise ValueError(f"Invalid ASTC data size: {len(image_data)} < {texture_size}")

    context, lock = get_astc_context(block_size)
    with lock:
        context.decompress(image_data[:texture_size], image, astc_encoder.ASTCSwizzle.from_str("RGBA"))
    assert image.data is not None, "Decompression failed, image data is None"

    return Image.frombytes("RGBA", (width, height), image.data, "raw", "RGBA")


ASTC_CONTEXTS: Dict[Tuple[int, int], Tuple[astc_encoder.ASTCContext, Lock]] = {}


def get_astc_context(block_size: tuple):
    """Get the ASTC context and its lock using the given `block_size`."""
    if block_size not in ASTC_CONTEXTS:
        config = astc_encoder.ASTCConfig(
            astc_encoder.ASTCProfile.LDR,
            *block_size,
            block_z=1,
            quality=100,
            flags=astc_encoder.ASTCConfigFlags.USE_DECODE_UNORM8,
        )
        context = astc_encoder.ASTCContext(config)
        lock = Lock()
        ASTC_CONTEXTS[block_size] = (context, lock)
    return ASTC_CONTEXTS[block_size]


def calculate_astc_compressed_size(width: int, height: int, block_size: tuple) -> int:
    """Calculate the size of the compressed data for ASTC."""
    # calculate the number of blocks
    block_count_x = (width + block_size[0] - 1) // block_size[0]
    block_count_y = (height + block_size[1] - 1) // block_size[1]
    # ignore depth for 2D textures
    # calculate the size of the compressed data
    return block_count_x * block_count_y * 16


def pvrtc(image_data: bytes, width: int, height: int, fmt: bool) -> Image.Image:
    image_data = texture2ddecoder.decode_pvrtc(image_data, width, height, fmt)
    return Image.frombytes("RGBA", (width, height), image_data, "raw", "BGRA")


def etc(image_data: bytes, width: int, height: int, fmt: str) -> Image.Image:
    if fmt == "ETC1":
        image_data = texture2ddecoder.decode_etc1(image_data, width, height)
    elif fmt == "ETC2_RGB":
        image_data = texture2ddecoder.decode_etc2(image_data, width, height)
    elif fmt == "ETC2_A1":
        image_data = texture2ddecoder.decode_etc2a1(image_data, width, height)
    elif fmt == "ETC2_A8":
        image_data = texture2ddecoder.decode_etc2a8(image_data, width, height)
    else:
        raise NotImplementedError(f"Unknown ETC mode: {fmt}")

    return Image.frombytes("RGBA", (width, height), image_data, "raw", "BGRA")


def eac(image_data: bytes, width: int, height: int, fmt: str) -> Image.Image:
    if fmt == "EAC_R":
        image_data = texture2ddecoder.decode_eacr(image_data, width, height)
    elif fmt == "EAC_R_SIGNED":
        image_data = texture2ddecoder.decode_eacr_signed(image_data, width, height)
    elif fmt == "EAC_RG":
        image_data = texture2ddecoder.decode_eacrg(image_data, width, height)
    elif fmt == "EAC_RG_SIGNED":
        image_data = texture2ddecoder.decode_eacrg_signed(image_data, width, height)
    else:
        raise NotImplementedError(f"Unknown EAC mode: {fmt}")

    return Image.frombytes("RGBA", (width, height), image_data, "raw", "BGRA")


def half(
    image_data: bytes,
    width: int,
    height: int,
    mode: str,
    codec: str,
    args,
    swap: Optional[tuple] = None,
) -> Image.Image:
    # convert half-float to int8
    stream = BytesIO(image_data)
    image_data = bytes(int(struct.unpack("e", stream.read(2))[0] * 256) for _ in range(width * height * len(codec)))
    return pillow(image_data, width, height, mode, codec, args, swap)


RG_PADDING_MAP = {
    "RGE": 16,
    "RGF": 32,
    "RG;16": 16,
    "RG;16s": 16,
    "RG;8s": 8,
}


def rg(image_data: bytes, width: int, height: int, mode: str, codec: str, args) -> Image.Image:
    # convert rg to rgb by adding in zeroes
    padding_size = RG_PADDING_MAP[codec]
    stream = BytesIO(image_data)
    padding = bytes(padding_size)
    rgb_data = b"".join(stream.read(padding_size * 2) + padding for _ in range(len(image_data) // (2 * padding_size)))
    if codec == "RGE":
        return half(rgb_data, width, height, mode, "RGB", args)
    else:
        return pillow(rgb_data, width, height, mode, codec.replace("RG", "RGB"), args)


def rgb9e5float(image_data: bytes, width: int, height: int) -> Image.Image:
    rgb = bytearray(width * height * 3)
    for i, (n,) in enumerate(struct.iter_unpack("<i", image_data)):
        scale = n >> 27 & 0x1F
        scalef = 2 ** (scale - 24)
        scaleb = scalef * 255.0
        b = (n >> 18 & 0x1FF) * scaleb
        g = (n >> 9 & 0x1FF) * scaleb
        r = (n & 0x1FF) * scaleb

        offset = i * 3
        rgb[offset : offset + 3] = [r, g, b]

    return Image.frombytes("RGB", (width, height), rgb, "raw", "RGB")


# Mapping TextureFormat -> (converter function, (additional args, ...))
CONV_TABLE: Dict[TF, Tuple[Callable[..., Image.Image], Tuple[Any, ...]]] = {
    TF.Alpha8: (pillow, ("RGBA", "raw", "A")),
    TF.ARGB4444: (pillow, ("RGBA", "raw", "RGBA;4B", (2, 1, 0, 3))),
    TF.RGB24: (pillow, ("RGB", "raw", "RGB")),
    TF.RGBA32: (pillow, ("RGBA", "raw", "RGBA")),
    TF.ARGB32: (pillow, ("RGBA", "raw", "ARGB")),
    TF.ARGBFloat: (pillow, ("RGBA", "raw", "RGBAF", (2, 1, 0, 3))),
    TF.RGB565: (pillow, ("RGB", "raw", "BGR;16")),
    TF.BGR24: (pillow, ("RGB", "raw", "BGR")),
    TF.R8: (pillow, ("RGB", "raw", "R")),
    TF.R16: (pillow, ("RGB", "raw", "R;16")),
    TF.RG16: (rg, ("RGB", "raw", "RG")),
    TF.DXT1: (pillow, ("RGBA", "bcn", 1)),
    TF.DXT3: (pillow, ("RGBA", "bcn", 2)),
    TF.DXT5: (pillow, ("RGBA", "bcn", 3)),
    TF.RGBA4444: (pillow, ("RGBA", "raw", "RGBA;4B", (3, 2, 1, 0))),
    TF.BGRA32: (pillow, ("RGBA", "raw", "BGRA")),
    TF.RHalf: (half, ("R", "raw", "R")),
    TF.RGHalf: (rg, ("RGB", "raw", "RGE")),
    TF.RGBAHalf: (half, ("RGB", "raw", "RGB")),
    TF.RFloat: (pillow, ("RGB", "raw", "RF")),
    TF.RGFloat: (rg, ("RGB", "raw", "RGF")),
    TF.RGBAFloat: (pillow, ("RGBA", "raw", "RGBAF")),
    # TF.YUY2: NotImplementedError("YUY2 not implemented"),
    TF.RGB9e5Float: (rgb9e5float, ()),
    TF.BC4: (pillow, ("L", "bcn", 4)),
    TF.BC5: (pillow, ("RGB", "bcn", 5)),
    TF.BC6H: (pillow, ("RGBA", "bcn", 6)),
    TF.BC7: (pillow, ("RGBA", "bcn", 7)),
    TF.DXT1Crunched: (pillow, ("RGBA", "bcn", 1)),
    TF.DXT5Crunched: (pillow, ("RGBA", "bcn", 3)),
    TF.PVRTC_RGB2: (pvrtc, (True,)),
    TF.PVRTC_RGBA2: (pvrtc, (True,)),
    TF.PVRTC_RGB4: (pvrtc, (False,)),
    TF.PVRTC_RGBA4: (pvrtc, (False,)),
    TF.ETC_RGB4: (etc, ("ETC1",)),
    TF.ATC_RGB4: (atc, (False,)),
    TF.ATC_RGBA8: (atc, (True,)),
    TF.EAC_R: (eac, ("EAC_R",)),
    TF.EAC_R_SIGNED: (eac, ("EAC_R_SIGNED",)),
    TF.EAC_RG: (eac, ("EAC_RG",)),
    TF.EAC_RG_SIGNED: (eac, ("EAC_RG_SIGNED",)),
    TF.ETC2_RGB: (etc, ("ETC2_RGB",)),
    TF.ETC2_RGBA1: (etc, ("ETC2_A1",)),
    TF.ETC2_RGBA8: (etc, ("ETC2_A8",)),
    TF.ASTC_RGB_4x4: (astc, ((4, 4),)),
    TF.ASTC_RGB_5x5: (astc, ((5, 5),)),
    TF.ASTC_RGB_6x6: (astc, ((6, 6),)),
    TF.ASTC_RGB_8x8: (astc, ((8, 8),)),
    TF.ASTC_RGB_10x10: (astc, ((10, 10),)),
    TF.ASTC_RGB_12x12: (astc, ((12, 12),)),
    TF.ASTC_RGBA_4x4: (astc, ((4, 4),)),
    TF.ASTC_RGBA_5x5: (astc, ((5, 5),)),
    TF.ASTC_RGBA_6x6: (astc, ((6, 6),)),
    TF.ASTC_RGBA_8x8: (astc, ((8, 8),)),
    TF.ASTC_RGBA_10x10: (astc, ((10, 10),)),
    TF.ASTC_RGBA_12x12: (astc, ((12, 12),)),
    TF.ETC_RGB4_3DS: (etc, ("ETC1",)),
    TF.ETC_RGBA8_3DS: (etc, ("ETC1",)),
    TF.ETC_RGB4Crunched: (etc, ("ETC1",)),
    TF.ETC2_RGBA8Crunched: (etc, ("ETC2_A8",)),
    TF.ASTC_HDR_4x4: (astc, ((4, 4),)),
    TF.ASTC_HDR_5x5: (astc, ((5, 5),)),
    TF.ASTC_HDR_6x6: (astc, ((6, 6),)),
    TF.ASTC_HDR_8x8: (astc, ((8, 8),)),
    TF.ASTC_HDR_10x10: (astc, ((10, 10),)),
    TF.ASTC_HDR_12x12: (astc, ((12, 12),)),
    TF.RG32: (rg, ("RGB", "raw", "RG;16")),
    TF.RGB48: (pillow, ("RGB", "raw", "RGB;16")),
    TF.RGBA64: (pillow, ("RGBA", "raw", "RGBA;16")),
    TF.R8_SIGNED: (pillow, ("R", "raw", "R;8s")),
    TF.RG16_SIGNED: (rg, ("RGB", "raw", "RG;8s")),
    TF.RGB24_SIGNED: (pillow, ("RGB", "raw", "RGB;8s")),
    TF.RGBA32_SIGNED: (pillow, ("RGBA", "raw", "RGBA;8s")),
    TF.R16_SIGNED: (pillow, ("R", "raw", "R;16s")),
    TF.RG32_SIGNED: (rg, ("RGB", "raw", "RG;16s")),
    TF.RGB48_SIGNED: (pillow, ("RGB", "raw", "RGB;16s")),
    TF.RGBA64_SIGNED: (pillow, ("RGBA", "raw", "RGBA;16s")),
}

# XBOX Swap Formats
XBOX_SWAP_FORMATS = [TF.RGB565, TF.DXT1, TF.DXT1Crunched, TF.DXT5, TF.DXT5Crunched]



from typing import List, Optional, Union

from UnityPy.classes.generated import (
    AABB,
    BlendShapeData,
    BoneInfluence,
    BoneWeights4,
    CompressedMesh,
    MeshBlendShape,
    MeshBlendShapeVertex,
    MinMaxAABB,
    NamedObject,
    StreamingInfo,
    SubMesh,
    VariableBoneCountWeights,
    VertexData,
)
from UnityPy.classes.math import ColorRGBA, Matrix4x4f, Vector2f, Vector3f, Vector4f

class Mesh(NamedObject):
    m_BindPose: List[Matrix4x4f]
    m_CompressedMesh: CompressedMesh
    m_IndexBuffer: List[int]
    m_LocalAABB: AABB
    m_MeshCompression: int
    m_MeshUsageFlags: int
    m_Name: str
    m_SubMeshes: List[SubMesh]
    m_BakedConvexCollisionMesh: Optional[List[int]] = None
    m_BakedTriangleCollisionMesh: Optional[List[int]] = None
    m_BoneNameHashes: Optional[List[int]] = None
    m_BonesAABB: Optional[List[MinMaxAABB]] = None
    m_CollisionTriangles: Optional[List[int]] = None
    m_CollisionVertexCount: Optional[int] = None
    m_Colors: Optional[List[ColorRGBA]] = None
    m_CookingOptions: Optional[int] = None
    m_IndexFormat: Optional[int] = None
    m_IsReadable: Optional[bool] = None
    m_KeepIndices: Optional[bool] = None
    m_KeepVertices: Optional[bool] = None
    m_MeshMetrics_0_: Optional[float] = None
    m_MeshMetrics_1_: Optional[float] = None
    m_Normals: Optional[List[Vector3f]] = None
    m_RootBoneNameHash: Optional[int] = None
    m_ShapeVertices: Optional[List[MeshBlendShapeVertex]] = None
    m_Shapes: Optional[Union[BlendShapeData, List[MeshBlendShape]]] = None
    m_Skin: Optional[Union[List[BoneInfluence], List[BoneWeights4]]] = None
    m_StreamCompression: Optional[int] = None
    m_StreamData: Optional[StreamingInfo] = None
    m_Tangents: Optional[List[Vector4f]] = None
    m_UV: Optional[List[Vector2f]] = None
    m_UV1: Optional[List[Vector2f]] = None
    m_Use16BitIndices: Optional[int] = None
    m_VariableBoneCountWeights: Optional[VariableBoneCountWeights] = None
    m_VertexData: Optional[VertexData] = None
    m_Vertices: Optional[List[Vector3f]] = None

    def export(self, format: str = "obj") -> str: ...



from ..generated import Sprite


def _Sprite_image(self: Sprite):
    from ...export import SpriteHelper

    return SpriteHelper.get_image_from_sprite(self)


Sprite.image = property(_Sprite_image)

__all__ = ("Sprite",)



from __future__ import annotations

import re
from sys import version_info as py_version_info
from typing import TYPE_CHECKING, Any, Optional, Union

from attrs import define

from .. import classes
from ..streams.EndianBinaryReader import EndianBinaryReader
from ..streams.EndianBinaryWriter import EndianBinaryWriter
from .TypeTreeNode import TypeTreeNode

Object = classes.Object
UnknownObject = classes.UnknownObject
PPtr = classes.PPtr

if TYPE_CHECKING:
    from ..files.SerializedFile import SerializedFile


try:
    from ..UnityPyBoost import read_typetree as read_typetree_boost
except ImportError:
    read_typetree_boost = None

kAlignBytes = 0x4000

FUNCTION_READ_MAP = {
    "SInt8": EndianBinaryReader.read_byte,
    "UInt8": EndianBinaryReader.read_u_byte,
    "char": EndianBinaryReader.read_u_byte,
    "short": EndianBinaryReader.read_short,
    "SInt16": EndianBinaryReader.read_short,
    "unsigned short": EndianBinaryReader.read_u_short,
    "UInt16": EndianBinaryReader.read_u_short,
    "int": EndianBinaryReader.read_int,
    "SInt32": EndianBinaryReader.read_int,
    "unsigned int": EndianBinaryReader.read_u_int,
    "UInt32": EndianBinaryReader.read_u_int,
    "Type*": EndianBinaryReader.read_u_int,
    "long long": EndianBinaryReader.read_long,
    "SInt64": EndianBinaryReader.read_long,
    "unsigned long long": EndianBinaryReader.read_u_long,
    "UInt64": EndianBinaryReader.read_u_long,
    "FileSize": EndianBinaryReader.read_u_long,
    "float": EndianBinaryReader.read_float,
    "double": EndianBinaryReader.read_double,
    "bool": EndianBinaryReader.read_boolean,
    "string": EndianBinaryReader.read_aligned_string,
    "TypelessData": EndianBinaryReader.read_byte_array,
}

FUNCTION_READ_MAP_ARRAY = {
    "SInt8": EndianBinaryReader.read_byte_array,
    "UInt8": EndianBinaryReader.read_u_byte_array,
    "char": EndianBinaryReader.read_u_byte_array,
    "short": EndianBinaryReader.read_short_array,
    "SInt16": EndianBinaryReader.read_short_array,
    "unsigned short": EndianBinaryReader.read_u_short_array,
    "UInt16": EndianBinaryReader.read_u_short_array,
    "int": EndianBinaryReader.read_int_array,
    "SInt32": EndianBinaryReader.read_int_array,
    "unsigned int": EndianBinaryReader.read_u_int_array,
    "UInt32": EndianBinaryReader.read_u_int_array,
    "Type*": EndianBinaryReader.read_u_int_array,
    "long long": EndianBinaryReader.read_long_array,
    "SInt64": EndianBinaryReader.read_long_array,
    "unsigned long long": EndianBinaryReader.read_u_long_array,
    "UInt64": EndianBinaryReader.read_u_long_array,
    "FileSize": EndianBinaryReader.read_u_long_array,
    "float": EndianBinaryReader.read_float_array,
    "double": EndianBinaryReader.read_double_array,
    "bool": EndianBinaryReader.read_boolean_array,
}


@define(slots=True)
class TypeTreeConfig:
    as_dict: bool
    assetsfile: "Optional[SerializedFile]" = None
    has_registry: bool = False

    def copy(self) -> TypeTreeConfig:
        return TypeTreeConfig(self.as_dict, self.assetsfile, self.has_registry)


def get_ref_type_node(ref_object: dict, assetfile: SerializedFile) -> Optional[TypeTreeNode]:
    typ = ref_object["type"]
    if isinstance(typ, dict):
        cls = typ["class"]
        ns = typ["ns"]
        asm = typ["asm"]
    else:
        cls = getattr(typ, "class")
        ns = typ.ns
        asm = typ.asm

    if not assetfile or not assetfile.ref_types:
        raise ValueError("SerializedFile has no ref_types")

    if cls == "":
        return None

    for ref_type in assetfile.ref_types:
        if cls == ref_type.m_ClassName and ns == ref_type.m_NameSpace and asm == ref_type.m_AssemblyName:
            return ref_type.node
    else:
        raise ValueError(f"Referenced type not found: {cls} {ns} {asm}")


if py_version_info >= (3, 14):
    from annotationlib import get_annotations as annotationlib_get_annotations

    def get_annotation_keys(clz) -> set[str]:
        return set(annotationlib_get_annotations(clz).keys())
else:

    def get_annotation_keys(clz) -> set[str]:
        return set(clz.__annotations__)


def read_typetree(
    root_node: TypeTreeNode,
    reader: EndianBinaryReader,
    as_dict: bool = True,
    byte_size: Optional[int] = None,
    check_read: bool = True,
    assetsfile: Optional[SerializedFile] = None,
) -> Union[dict[str, Any], Object]:
    """Reads the typetree of the object contained in the reader via the node list.

    Parameters
    ----------
    nodes : list
        List of nodes/nodes
    reader : EndianBinaryReader
        Reader of the object to be parsed

    Returns
    -------
    dict | objects.Object
        The parsed typtree
    """
    bytes_read: int
    if byte_size and read_typetree_boost:
        data = reader.read_bytes(byte_size)
        obj, bytes_read = read_typetree_boost(data, root_node, reader.endian, as_dict, assetsfile, classes)
    else:
        pos = reader.Position
        config = TypeTreeConfig(as_dict, assetsfile, False)
        obj = read_value(root_node, reader, config)
        bytes_read = reader.Position - pos

    if check_read and bytes_read != byte_size:
        raise ValueError(f"Expected to read {byte_size} bytes, but only read {bytes_read} bytes")

    return obj


def write_typetree(
    value: Union[dict[str, Any], Object],
    root_node: TypeTreeNode,
    writer: EndianBinaryWriter,
    assetsfile: Optional[SerializedFile] = None,
) -> None:
    """Writes the typetree of the object contained in the reader via the node list.

    Parameters
    ----------
    value: dict | objects.Object
        The object to be written
    nodes : list
        List of nodes/nodes
    writer : EndianBinaryWriter
        Writer of the object to be parsed
    """
    config = TypeTreeConfig(isinstance(value, dict), assetsfile, False)
    return write_value(value, root_node, writer, config)


def read_value(
    node: TypeTreeNode,
    reader: EndianBinaryReader,
    config: TypeTreeConfig,
) -> Any:
    # print(reader.Position, node.m_Name, node.m_Type, node.m_MetaFlag)
    align = metaflag_is_aligned(node.m_MetaFlag)

    func = FUNCTION_READ_MAP.get(node.m_Type)
    if func:
        value = func(reader)
    elif node.m_Type == "pair":
        first = read_value(node.m_Children[0], reader, config)
        second = read_value(node.m_Children[1], reader, config)
        value = (first, second)
    elif node.m_Type == "ReferencedObject":
        value = {}
        for child in node.m_Children:
            if child.m_Type == "ReferencedObjectData":
                ref_type_nodes = get_ref_type_node(value, config.assetsfile)
                if ref_type_nodes is None:
                    continue
                value[child.m_Name] = read_value(ref_type_nodes, reader, config)
            else:
                value[child.m_Name] = read_value(child, reader, config)
    # Vector
    elif node.m_Children and node.m_Children[0].m_Type == "Array":
        if metaflag_is_aligned(node.m_Children[0].m_MetaFlag):
            align = True

        # size = read_value(node.m_Children[0].m_Children[0], reader, as_dict)
        size = reader.read_int()
        if size < 0:
            raise ValueError("Negative length read from TypeTree")
        subtype = node.m_Children[0].m_Children[1]
        if metaflag_is_aligned(subtype.m_MetaFlag):
            value = read_value_array(subtype, reader, config, size)
        else:
            value = [read_value(subtype, reader, config) for _ in range(size)]

    else:  # Class
        value = {}
        for child in node.m_Children:
            if child.m_Type == "ManagedReferencesRegistry":
                if config.has_registry:
                    continue
                else:
                    config = config.copy()
                    config.has_registry = True
            value[child.m_Name if config.as_dict else child._clean_name] = read_value(child, reader, config)

        if not config.as_dict:
            if node.m_Type.startswith("PPtr<"):
                value = PPtr[Any](
                    assetsfile=config.assetsfile,
                    m_FileID=value["m_FileID"],
                    m_PathID=value["m_PathID"],
                )
            else:
                clz = getattr(classes, node.m_Type, UnknownObject)
                try:
                    value = clz(**value)
                except TypeError:
                    keys = set(value.keys())
                    annotation_keys = get_annotation_keys(clz)
                    missing_keys = annotation_keys - keys
                    if clz is UnknownObject or missing_keys:
                        value = UnknownObject(node, **value)
                    else:
                        extra_keys = keys - annotation_keys
                        if extra_keys:
                            instance = clz(**{key: value[key] for key in annotation_keys})
                            for key in extra_keys:
                                setattr(instance, key, value[key])
                            value = instance
                        else:
                            value = UnknownObject(**value)

    if align:
        reader.align_stream()

    return value


def read_value_array(
    node: TypeTreeNode,
    reader: EndianBinaryReader,
    config: TypeTreeConfig,
    size: int,
) -> Any:
    align = metaflag_is_aligned(node.m_MetaFlag)

    func = FUNCTION_READ_MAP_ARRAY.get(node.m_Type)
    if func:
        value = func(reader, size)
    elif node.m_Type == "string":
        value = [reader.read_aligned_string() for _ in range(size)]
    elif node.m_Type == "TypelessData":
        value = [reader.read_bytes() for _ in range(size)]
    elif node.m_Type == "pair":
        key_node = node.m_Children[0]
        value_node = node.m_Children[1]

        key_func = FUNCTION_READ_MAP.get(
            key_node.m_Type,
            lambda reader: read_value(key_node, reader, config),
        )
        value_func = FUNCTION_READ_MAP.get(
            value_node.m_Type,
            lambda reader: read_value(value_node, reader, config),
        )
        value = [(key_func(reader), value_func(reader)) for _ in range(size)]
    elif node.m_Type == "ReferencedObject":
        value = [None] * size
        for i in range(size):
            item = {}
            for child in node.m_Children:
                if child.m_Type == "ReferencedObjectData":
                    ref_type_nodes = get_ref_type_node(item, config.assetsfile)
                    item[child.m_Name] = read_value(ref_type_nodes, reader, config)
                else:
                    item[child.m_Name] = read_value(child, reader, config)
            value[i] = item
    # Vector
    elif node.m_Children and node.m_Children[0].m_Type == "Array":
        if metaflag_is_aligned(node.m_Children[0].m_MetaFlag):
            align = True
        subtype = node.m_Children[0].m_Children[1]
        if metaflag_is_aligned(subtype.m_MetaFlag):
            value = [read_value_array(subtype, reader, config, reader.read_int()) for _ in range(size)]
        else:
            value = [[read_value(subtype, reader, config) for _ in range(reader.read_int())] for _ in range(size)]
    else:  # Class
        if config.as_dict:
            value = [
                {child.m_Name: read_value(child, reader, config) for child in node.m_Children} for _ in range(size)
            ]
        elif node.m_Type.startswith("PPtr<"):
            value = [
                PPtr[Any](
                    assetsfile=config.assetsfile,
                    **{child.m_Name: read_value(child, reader, config) for child in node.m_Children},
                )
                for _ in range(size)
            ]
        else:
            clz = getattr(
                classes,
                node.m_Type,
                UnknownObject,
            )
            keys = set(child._clean_name for child in node.m_Children)
            annotation_keys = get_annotation_keys(clz)
            missing_keys = annotation_keys - keys
            extra_keys = keys - annotation_keys
            if missing_keys or clz is UnknownObject:
                value = [
                    UnknownObject(
                        node,
                        **{child._clean_name: read_value(child, reader, config) for child in node.m_Children},
                    )
                    for _ in range(size)
                ]
            elif extra_keys:
                value = [None] * size
                for i in range(size):
                    value_i_d = {child._clean_name: read_value(child, reader, config) for child in node.m_Children}
                    value_i = clz(**{key: value for key, value in value_i_d.items() if key in annotation_keys})
                    for key in extra_keys:
                        setattr(value_i, key, value_i_d[key])
                    value[i] = value_i
            else:
                value = [
                    clz(**{child._clean_name: read_value(child, reader, config) for child in node.m_Children})
                    for _ in range(size)
                ]

    if align:
        reader.align_stream()
    return value


def metaflag_is_aligned(meta_flag: int | None) -> bool:
    return ((meta_flag or 0) & kAlignBytes) != 0


def clean_name(name: str) -> str:
    if name.startswith("(int&)"):
        name = name[6:]
    if name.endswith("?"):
        name = name[:-1]
    name = re.sub(r"[ \.:\-\[\]]", "_", name)
    if name in ["pass", "from"]:
        name += "_"
    if name[0].isdigit():
        name = f"x{name}"
    return name


FUNCTION_WRITE_MAP = {
    "SInt8": EndianBinaryWriter.write_byte,
    "UInt8": EndianBinaryWriter.write_u_byte,
    "char": EndianBinaryWriter.write_u_byte,
    "short": EndianBinaryWriter.write_short,
    "SInt16": EndianBinaryWriter.write_short,
    "unsigned short": EndianBinaryWriter.write_u_short,
    "UInt16": EndianBinaryWriter.write_u_short,
    "int": EndianBinaryWriter.write_int,
    "SInt32": EndianBinaryWriter.write_int,
    "unsigned int": EndianBinaryWriter.write_u_int,
    "UInt32": EndianBinaryWriter.write_u_int,
    "Type*": EndianBinaryWriter.write_u_int,
    "long long": EndianBinaryWriter.write_long,
    "SInt64": EndianBinaryWriter.write_long,
    "unsigned long long": EndianBinaryWriter.write_u_long,
    "UInt64": EndianBinaryWriter.write_u_long,
    "FileSize": EndianBinaryWriter.write_u_long,
    "float": EndianBinaryWriter.write_float,
    "double": EndianBinaryWriter.write_double,
    "bool": EndianBinaryWriter.write_boolean,
    "string": EndianBinaryWriter.write_aligned_string,
    "TypelessData": EndianBinaryWriter.write_byte_array,
}


def write_value(
    value: Union[dict[str, Any], Object],
    node: TypeTreeNode,
    writer: EndianBinaryWriter,
    config: TypeTreeConfig,
) -> None:
    # print(reader.Position, node.m_Name, node.m_Type, node.m_MetaFlag)
    align = metaflag_is_aligned(node.m_MetaFlag)

    func = FUNCTION_WRITE_MAP.get(node.m_Type)
    if func:
        value = func(writer, value)
    elif node.m_Type == "pair":
        write_value(value[0], node.m_Children[0], writer, config)
        write_value(value[1], node.m_Children[1], writer, config)
    elif node.m_Type == "ReferencedObject":
        for child in node.m_Children:
            if child.m_Type == "ReferencedObjectData":
                ref_type_nodes = get_ref_type_node(value, config.assetsfile)
                write_value(value[child.m_Name], ref_type_nodes, writer, config)
            else:
                write_value(value[child.m_Name], child, writer, config)
    elif node.m_Children and node.m_Children[0].m_Type == "Array":
        if metaflag_is_aligned(node.m_Children[0].m_MetaFlag):
            align = True

        writer.write_int(len(value))
        subtype = node.m_Children[0].m_Children[1]
        [write_value(sub_value, subtype, writer, config) for sub_value in value]

    else:  # Class
        if isinstance(value, dict):
            for child in node.m_Children:
                if child.m_Type == "ManagedReferencesRegistry":
                    if config.has_registry:
                        continue
                    else:
                        config = config.copy()
                        config.has_registry = True
                write_value(value[child.m_Name], child, writer, config)
        else:
            for child in node.m_Children:
                if child.m_Type == "ManagedReferencesRegistry":
                    if config.has_registry:
                        continue
                    else:
                        config = config.copy()
                        config.has_registry = True
                write_value(getattr(value, child._clean_name), child, writer, config)

    if align:
        writer.align_stream()



from .Audio import AUDIO_TYPE_EXTEMSION, AudioCompressionFormat, AudioType
from .BuildTarget import BuildTarget
from .BundleFile import ArchiveFlags, ArchiveFlagsOld, CompressionFlags
from .ClassIDType import ClassIDType
from .CommonString import CommonString
from .FileType import FileType
from .GfxPrimitiveType import GfxPrimitiveType
from .GraphicsFormat import GraphicsFormat
from .PassType import PassType
from .SerializedPropertyType import SerializedPropertyType
from .ShaderCompilerPlatform import ShaderCompilerPlatform
from .ShaderGpuProgramType import ShaderGpuProgramType
from .SpriteMeshType import SpriteMeshType
from .SpritePackingMode import SpritePackingMode
from .SpritePackingRotation import SpritePackingRotation
from .TextureDimension import TextureDimension
from .TextureFormat import TextureFormat

__all__ = [
    "AUDIO_TYPE_EXTEMSION",
    "AudioCompressionFormat",
    "AudioType",
    "BuildTarget",
    "ArchiveFlags",
    "ArchiveFlagsOld",
    "CompressionFlags",
    "ClassIDType",
    "CommonString",
    "FileType",
    "GfxPrimitiveType",
    "GraphicsFormat",
    "PassType",
    "SerializedPropertyType",
    "ShaderCompilerPlatform",
    "ShaderGpuProgramType",
    "SpriteMeshType",
    "SpritePackingMode",
    "SpritePackingRotation",
    "TextureDimension",
    "TextureFormat",
]



import io
import ntpath
import os
import re
from typing import TYPE_CHECKING, BinaryIO, Callable, Dict, List, Optional, Union, cast
from zipfile import ZipFile

from fsspec import AbstractFileSystem
from fsspec.implementations.local import LocalFileSystem

from .enums import FileType
from .files import BundleFile, File, ObjectReader, SerializedFile, WebFile
from .helpers.ContainerHelper import ContainerHelper
from .helpers.ImportHelper import (
    FileSourceType,
    check_file_type,
    find_sensitive_path,
    parse_file,
)
from .streams import EndianBinaryReader

if TYPE_CHECKING:
    from UnityPy.helpers.TypeTreeGenerator import TypeTreeGenerator

reSplit = re.compile(r"(.*?([^\/\\]+?))\.split\d+")


class Environment:
    files: Dict[str, Union[SerializedFile, BundleFile, WebFile, EndianBinaryReader]]
    cabs: Dict[str, Union[SerializedFile, EndianBinaryReader]]
    path: str
    local_files: List[str]
    local_files_simple: List[str]
    typetree_generator: Optional["TypeTreeGenerator"] = None

    def __init__(self, *args: FileSourceType, fs: Optional[AbstractFileSystem] = None, path: Optional[str] = None):
        self.files = {}
        self.cabs = {}
        self.fs = fs or LocalFileSystem()
        self.local_files = []
        self.local_files_simple = []

        if path is None:
            # if no path is given, use the current working directory
            if isinstance(self.fs, LocalFileSystem):
                self.path = os.getcwd()
            else:
                self.path = ""
        else:
            self.path = path

        if args:
            for arg in args:
                if isinstance(arg, str):
                    if self.fs.isfile(arg):
                        if ntpath.splitext(arg)[-1] in [".apk", ".zip"]:
                            self.load_zip_file(arg)
                        else:
                            self.path = ntpath.dirname(arg)
                            if reSplit.match(arg):
                                self.load_files([arg])
                            else:
                                self.load_file(arg)
                    elif self.fs.isdir(arg):
                        self.path = arg
                        self.load_folder(arg)
                else:
                    self.load_file(file=arg)

        if len(self.files) == 1:
            self.file = list(self.files.values())[0]

    def load_files(self, files: List[str]):
        """Loads all files (list) into the Environment and merges .split files for common usage."""
        self.load_assets(files, lambda x: open(x, "rb"))

    def load_folder(self, path: str):
        """Loads all files in the given path and its subdirs into the Environment."""
        self.load_files([self.fs.sep.join([root, f]) for root, dirs, files in self.fs.walk(path) for f in files])

    def load(self, files: List[str]):
        """Loads all files into the Environment."""
        self.files.update(
            {ntpath.basename(f): self.load_file(self.fs.open(f, "rb"), self, f) for f in files if self.fs.exists(f)}
        )

    def _load_split_file(self, basename: str) -> bytes:
        file: List[bytes] = []
        for i in range(0, 999):
            item = f"{basename}.split{i}"
            if self.fs.exists(item):
                with self.fs.open(item, "rb") as f:
                    file.append(f.read())  # type: ignore
            elif i:
                break
        return b"".join(file)

    def load_file(
        self,
        file: FileSourceType,
        parent: Optional[Union["Environment", File]] = None,
        name: Optional[str] = None,
        is_dependency: bool = False,
    ):
        if not parent:
            parent = self

        if isinstance(file, str):
            split_match = reSplit.match(file)
            if split_match:
                basepath, _basename = split_match.groups()
                assert isinstance(basepath, str)
                name = basepath
                file = self._load_split_file(basepath)
            else:
                name = file
                if not os.path.exists(file):
                    # relative paths are in the asset directory, not the cwd
                    if not os.path.isabs(file):
                        file = os.path.join(self.path, file)
                    # for dependency loading of split files
                    if os.path.exists(f"{file}.split0"):
                        file = self._load_split_file(file)
                    # Unity paths are case insensitive,
                    # so we need to find "Resources/Foo.asset" when the record says "resources/foo.asset"
                    elif not os.path.exists(file):
                        file_path = find_sensitive_path(self.path, file)
                        if file_path:
                            file = file_path
                        else:
                            return None
                            # raise FileNotFoundError(f"File {file} not found in {self.path}")

                if isinstance(file, str):
                    file = self.fs.open(file, "rb")

        typ, reader = check_file_type(file)

        stream_name = (
            name
            if name
            else getattr(
                file,
                "name",
                str(file.__hash__()) if hasattr(file, "__hash__") else "",  # type: ignore
            )
        )

        if typ == FileType.ZIP:
            f = self.load_zip_file(file)
        else:
            f = parse_file(reader, self, name=stream_name, typ=typ, is_dependency=is_dependency)

        if isinstance(f, (SerializedFile, EndianBinaryReader)):
            self.register_cab(stream_name, f)

        self.files[stream_name] = f
        return f

    def load_zip_file(self, value):
        if isinstance(value, str) and self.fs.exists(value):
            buffer = cast(io.BufferedReader, self.fs.open(value, "rb"))
        elif isinstance(value, (bytes, bytearray, memoryview)):
            buffer = io.BytesIO(value)
        elif isinstance(value, (io.BufferedReader, io.BufferedIOBase)):
            buffer = value
        else:
            raise TypeError("Unsupported type for loading zip file")

        z = ZipFile(buffer)
        self.load_assets(z.namelist(), lambda x: z.open(x, "r"))  # type: ignore
        z.close()

    def save(self, pack="none", out_path="output"):
        """Saves all changed assets.
        Mark assets as changed using `.mark_changed()`.
        pack = "none" (default) or "lz4"
        """
        for fname, fitem in self.files.items():
            if getattr(fitem, "is_changed", False):
                with open(self.fs.sep.join([out_path, ntpath.basename(fname)]), "wb") as out:
                    out.write(fitem.save(packer=pack))

    @property
    def objects(self) -> List[ObjectReader]:
        """Returns a list of all objects in the Environment."""

        def search(item):
            ret = []
            if not isinstance(item, Environment) and getattr(item, "objects", None):
                # serialized file
                if getattr(item, "is_dependency", False):
                    return []
                return [val for val in item.objects.values()]

            elif getattr(item, "files", None):  # WebBundle and BundleFile
                # bundle
                for sub_item in item.files.values():
                    ret.extend(search(sub_item))
                return ret

            return ret

        return search(self)

    @property
    def container(self) -> ContainerHelper:
        """Returns a dictionary of all objects in the Environment."""
        container = []
        for f in self.cabs.values():
            if isinstance(f, SerializedFile) and not f.is_dependency:
                container.extend(f.container.container)

        return ContainerHelper(container)

    @property
    def assets(self) -> list:
        """
        Lists all assets / SerializedFiles within this environment.
        """

        def gen_all_asset_files(file, ret: Optional[list] = None):
            if ret is None:
                ret = []

            for f in getattr(file, "files", {}).values():
                if getattr(f, "is_dependency", False):
                    continue
                if isinstance(f, SerializedFile):
                    ret.append(f)
                else:
                    gen_all_asset_files(f, ret)
            return ret

        return gen_all_asset_files(self)

    def get(self, key: str, default=None):
        return getattr(self, key, default)

    def register_cab(self, name: str, item: Union[SerializedFile, EndianBinaryReader]) -> None:
        """
        Registers a cab file.

        Parameters
        ----------
        name : str
            The name of the cab file.
        item : File
            The file to register.
        """
        self.cabs[simplify_name(name)] = item

    def get_cab(self, name: str) -> Union[SerializedFile, EndianBinaryReader, None]:
        """
        Returns the cab file with the given name.

        Parameters
        ----------
        name : str
            The name of the cab file.

        Returns
        -------
        File
            The cab file.
        """
        return self.cabs.get(simplify_name(name), None)

    def load_assets(self, assets: List[str], open_f: Callable[[str], BinaryIO]):
        """
        Load all assets from a list of files via the given open_f function.

        Parameters
        ----------
        assets : List[str]
            List of files to load.
        open_f : Callable[[str], io.IOBase]
            Function to open the files.
            The function takes a file path and returns an io.IOBase object.
        """
        split_files = []
        for path in assets:
            splitMatch = reSplit.match(path)
            if splitMatch:
                basepath, _basename = splitMatch.groups()

                if basepath in split_files:
                    continue

                split_files.append(basepath)
                data = self._load_split_file(basepath)
                path = basepath
            else:
                data = open_f(path).read()
            self.load_file(data, name=path)

    def find_file(self, name: str, is_dependency: bool = True) -> Union[File, None]:
        """
        Finds a file in the environment.

        Parameters
        ----------
        name : str
            The name of the file.
        is_dependency : bool
            Whether the file is a dependency.

        Returns
        -------
        File | None
            The file if it was found, otherwise None.
        """
        simple_name = simplify_name(name)
        cab = self.get_cab(simple_name)
        if cab:
            return cab
        fp = self.fs.sep.join([self.path, name])
        if self.fs.exists(fp):
            return self.load_file(fp, name=name, is_dependency=is_dependency)

        if len(self.local_files) == 0 and self.path:
            for root, _, files in self.fs.walk(self.path):
                for f in files:
                    self.local_files.append(self.fs.sep.join([root, f]))
                    self.local_files_simple.append(self.fs.sep.join([root, simplify_name(f)]))

        if name in self.local_files:
            fp = name
        elif simple_name in self.local_files_simple:
            fp = self.local_files[self.local_files_simple.index(simple_name)]
        else:
            fp = next((f for f in self.local_files if f.endswith(name)), None)
            if not fp:
                fp = next(
                    (f for f in self.local_files_simple if f.endswith(simple_name)),
                    None,
                )
            if not fp:
                raise FileNotFoundError(f"File {name} not found in {self.path}")

        return self.load_file(fp, name=name, is_dependency=is_dependency)


def simplify_name(name: str) -> str:
    """Simplifies a name by:
    - removing the extension
    - removing the path
    - converting to lowercase
    """
    return ntpath.basename(name).lower()



__version__ = "1.24.2"

from .environment import Environment as Environment
from .helpers.ArchiveStorageManager import (
    set_assetbundle_decrypt_key as set_assetbundle_decrypt_key,
)

load = Environment

# backward compatibility
AssetsManager = Environment



from ...enums import AUDIO_TYPE_EXTEMSION
from ..generated import AudioClip


def _AudioClip_extension(self: AudioClip) -> str:
    return AUDIO_TYPE_EXTEMSION.get(self.m_CompressionFormat, ".audioclip")


def _AudioClip_samples(self: AudioClip) -> dict:
    from ...export import AudioClipConverter

    return AudioClipConverter.extract_audioclip_samples(self)


AudioClip.extension = property(_AudioClip_extension)
AudioClip.samples = property(_AudioClip_samples)

__all__ = ("AudioClip",)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, List, Literal, Optional, Tuple, Union

if TYPE_CHECKING:
    from .classes import Object
    from .files.SerializedFile import SerializedFile

def unpack_vertexdata(
    data: Union[bytes, bytearray],
    component_byte_size: int,
    vertex_count: int,
    stream_offset: int,
    stream_stride: int,
    channel_offset: int,
    channel_dimension: int,
    swap: bool,
) -> bytes: ...
def read_typetree(
    data: Union[bytes, bytearray],
    node: TypeTreeNode,
    endian: Literal["<", ">"],
    as_dict: bool,
    assetsfile: SerializedFile,
    classes: dict,
) -> Tuple[Union[dict[str, Any], Object], int]: ...

class TypeTreeNode:
    m_Level: int
    m_Type: str
    m_Name: str
    m_ByteSize: int
    m_Version: int
    m_Children: List[TypeTreeNode]
    m_TypeFlags: Optional[int] = None
    m_VariableCount: Optional[int] = None
    m_Index: Optional[int] = None
    m_MetaFlag: Optional[int] = None
    m_RefTypeHash: Optional[int] = None
    _clean_name: str

def decrypt_block(
    index_bytes: Union[bytes, bytearray],
    substitute_bytes: Union[bytes, bytearray],
    data: Union[bytes, bytearray],
    index: int,
) -> bytes: ...



from typing import Optional

from ..helpers import CompressionHelper
from ..streams import EndianBinaryReader, EndianBinaryWriter
from . import File


class WebFile(File.File):
    """A package which can hold other WebFiles, Bundles and SerialiedFiles.
    It may be compressed via gzip or brotli.

    files -- list of all files in the WebFile
    """

    def __init__(self, reader: EndianBinaryReader, parent: File, name=None, **kwargs):
        """Constructor Method"""
        super().__init__(parent=parent, name=name, **kwargs)

        # check compression
        magic = reader.read_bytes(2)
        reader.Position = 0

        if magic == CompressionHelper.GZIP_MAGIC:
            self.packer = "gzip"
            data = CompressionHelper.decompress_gzip(reader.bytes)
            reader = EndianBinaryReader(data, endian="<")
        else:
            reader.Position = 0x20
            magic = reader.read_bytes(6)
            reader.Position = 0
            if CompressionHelper.BROTLI_MAGIC == magic:
                self.packer = "brotli"
                data = CompressionHelper.decompress_brotli(reader.bytes)
                reader = EndianBinaryReader(data, endian="<")
            else:
                self.packer = "none"
                reader.endian = "<"

        # signature check
        signature = reader.read_string_to_null()
        if not signature.startswith(("UnityWebData", "TuanjieWebData")):
            raise ValueError(f"Invalid WebFile signature: {signature!r}. Expected 'UnityWebData' or 'TuanjieWebData'.")
        self.signature = signature

        # read header -> contains file headers
        head_length = reader.read_int()

        files = []
        while reader.Position < head_length:
            offset = reader.read_int()
            length = reader.read_int()
            path_length = reader.read_int()
            name = bytes(reader.read_bytes(path_length)).decode("utf-8")
            files.append(File.DirectoryInfo(name, offset, length))

        self.read_files(reader, files)

    def save(
        self,
        files: Optional[dict] = None,
        packer: str = "none",
        signature: str = "UnityWebData1.0",
    ) -> bytes:
        # solve defaults
        if not files:
            files = self.files
        if not packer:
            packer = self.packer

        # get raw data
        files = {name: f.bytes if isinstance(f, EndianBinaryReader) else f.save() for name, f in files.items()}

        # create writer
        writer = EndianBinaryWriter(endian="<")
        # signature
        writer.write_string_to_null(signature)

        # data offset
        offset = sum(
            [
                writer.Position,  # signature
                sum(len(path.encode("utf-8")) for path in files.keys()),  # path of each file
                4 * 3 * len(files),  # 3 ints per file
                4,  # offset int
            ]
        )

        writer.write_int(offset)

        # 1. file headers
        for name, data in files.items():
            # offset
            writer.write_int(offset)
            # length
            length = len(data)
            writer.write_int(length)
            offset += length
            # path
            enc_path = name.encode("utf-8")
            writer.write_int(len(enc_path))
            writer.write(enc_path)

        # 2. file data
        for data in files.values():
            writer.write(data)

        if packer == "gzip":
            return CompressionHelper.compress_gzip(writer.bytes)
        elif packer == "brotli":
            return CompressionHelper.compress_brotli(writer.bytes)
        else:
            return writer.bytes



from .BundleFile import BundleFile
from .File import DirectoryInfo, File
from .ObjectReader import ObjectReader
from .SerializedFile import SerializedFile
from .WebFile import WebFile

__all__ = [
    "BundleFile",
    "DirectoryInfo",
    "File",
    "ObjectReader",
    "SerializedFile",
    "WebFile",
]



from __future__ import annotations

import math
import struct
from typing import TYPE_CHECKING, List, Optional, Tuple, Union, cast

from ..classes.generated import (
    ChannelInfo,
    Mesh,
    SpriteRenderData,
    StreamInfo,
    Vector2f,
    Vector3f,
    Vector4f,
)
from ..enums.MeshTopology import MeshTopology
from ..enums.VertexFormat import (
    VERTEX_CHANNEL_FORMAT_STRUCT_TYPE_MAP,
    VERTEX_FORMAT_2017_STRUCT_TYPE_MAP,
    VERTEX_FORMAT_STRUCT_TYPE_MAP,
    VertexChannelFormat,
    VertexFormat,
    VertexFormat2017,
)
from .PackedBitVector import unpack_floats, unpack_ints
from .ResourceReader import get_resource_data

try:
    from UnityPy import UnityPyBoost
except ImportError:
    UnityPyBoost = None

Tuple2f = Tuple[float, float]
Tuple3f = Tuple[float, float, float]
Tuple4f = Tuple[float, float, float, float]


def vector_list_to_tuples(
    data: Union[List[Vector2f], List[Vector3f], List[Vector4f]],
) -> List[tuple]:
    if isinstance(data[0], Vector2f):
        return [(v.x, v.y) for v in data]
    elif isinstance(data[0], Vector3f):
        if TYPE_CHECKING:
            data = cast(List[Vector3f], data)
        return [(v.x, v.y, v.z) for v in data]
    elif isinstance(data[0], Vector4f):
        if TYPE_CHECKING:
            data = cast(List[Vector4f], data)
        return [(v.x, v.y, v.z, v.w) for v in data]
    else:
        raise ValueError("Unknown vector type")


def lists_to_tuples(data: List[list]) -> List[tuple]:
    return [tuple(v) for v in data]


def zeros(m: int, n: int) -> List[list]:
    return [[0] * n for _ in range(m)]


def normalize(*vector: float) -> Tuple[float, ...]:
    length = math.sqrt(sum(v**2 for v in vector))
    if length > 0.00001:
        inv_norm = 1.0 / length
        return tuple(v * inv_norm for v in vector)
    return (0,) * len(vector)


class MeshHandler:
    src: Union[Mesh, SpriteRenderData]
    endianess: str = "<"
    version: Tuple[int, int, int, int]
    m_VertexCount: int = 0
    m_Vertices: Optional[List[Tuple3f]] = None
    # normals can be stored as Tuple4f,
    # in such cases the 4th dimension is always 0 and can be discarded
    m_Normals: Optional[Union[List[Tuple3f], List[Tuple4f]]] = None
    m_Colors: Optional[List[Tuple4f]] = None
    m_UV0: Optional[List[Tuple2f]] = None
    m_UV1: Optional[List[Tuple2f]] = None
    m_UV2: Optional[List[Tuple2f]] = None
    m_UV3: Optional[List[Tuple2f]] = None
    m_UV4: Optional[List[Tuple2f]] = None
    m_UV5: Optional[List[Tuple2f]] = None
    m_UV6: Optional[List[Tuple2f]] = None
    m_UV7: Optional[List[Tuple2f]] = None
    m_Tangents: Optional[List[Tuple4f]] = None
    m_BoneIndices: Optional[List[Tuple[int, int, int, int]]] = None
    m_BoneWeights: Optional[List[Tuple4f]] = None
    m_IndexBuffer: Optional[List[int]] = None
    m_Use16BitIndices: bool = True

    def __init__(
        self,
        src: Union[Mesh, SpriteRenderData],
        version: Optional[Tuple[int, int, int, int]] = None,
        endianess: str = "<",
    ):
        self.src = src
        self.endianess = endianess
        if version is not None:
            self.version = version
        elif not isinstance(src, SpriteRenderData) and src.object_reader is not None:
            self.version = src.object_reader.version
        else:
            raise ValueError("No version provided and no object reader found")

    def process(self):
        mesh = self.src
        vertex_data = mesh.m_VertexData
        assert vertex_data is not None

        m_Channels: list[ChannelInfo]
        m_Streams: list[StreamInfo]

        if self.version[0] < 4:
            assert (
                vertex_data.m_Streams_0_ is not None
                and vertex_data.m_Streams_1_ is not None
                and vertex_data.m_Streams_2_ is not None
                and vertex_data.m_Streams_3_ is not None
            )
            m_Streams = [
                vertex_data.m_Streams_0_,
                vertex_data.m_Streams_1_,
                vertex_data.m_Streams_2_,
                vertex_data.m_Streams_3_,
            ]
            assert all(stream is not None for stream in m_Streams)
            m_Channels = self.get_channels(m_Streams)
        elif self.version[0] == 4:
            assert vertex_data.m_Streams is not None and vertex_data.m_Channels is not None
            m_Streams = vertex_data.m_Streams
            m_Channels = vertex_data.m_Channels
        else:
            assert vertex_data.m_Channels is not None
            m_Channels = vertex_data.m_Channels
            m_Streams = self.get_streams(m_Channels, vertex_data.m_VertexCount)

        if (
            isinstance(mesh, Mesh) and mesh.m_StreamData and mesh.m_StreamData.path
            # and mesh.m_VertexData
            # and mesh.m_VertexData.m_VertexCount
        ):
            stream_data = mesh.m_StreamData
            assert mesh.object_reader, "No object reader assigned to the input Mesh!"
            data = get_resource_data(
                stream_data.path,
                mesh.object_reader.assets_file,
                stream_data.offset,
                stream_data.size,
            )
            vertex_data.m_DataSize = data

        # try to copy data directly from mesh
        if isinstance(mesh, Mesh):
            if mesh.m_Use16BitIndices is not None:
                self.m_Use16BitIndices = bool(mesh.m_Use16BitIndices)
            elif (
                self.version >= (2017, 4)
                or
                # version == (2017, 3, 1) & patched - px string
                self.version[:2] == (2017, 3)
                and mesh.m_MeshCompression == 0
            ):
                self.m_Use16BitIndices = mesh.m_IndexFormat == 0
            self.copy_from_mesh()
        elif isinstance(mesh, SpriteRenderData):
            self.copy_from_spriterenderdata()
        else:
            raise ValueError(f"Unknown mesh type {type(mesh)}")

        if self.m_IndexBuffer:
            raw_indices = bytes(self.m_IndexBuffer)
            if self.m_Use16BitIndices:
                char = "H"
                index_size = 2
            else:
                char = "I"
                index_size = 4

            self.m_IndexBuffer = cast(
                List[int],
                struct.unpack(f"<{len(raw_indices) // index_size}{char}", raw_indices),
            )

        if self.version >= (3, 5):
            self.read_vertex_data(m_Channels, m_Streams)

        if isinstance(mesh, Mesh) and self.version >= (2, 6):
            self.decompress_compressed_mesh()

        if self.m_VertexCount == 0 and self.m_Vertices:
            self.m_VertexCount = len(self.m_Vertices)

    def copy_from_mesh(self):
        """Copy data from mesh to handler if it's not already set."""
        mesh = self.src
        if TYPE_CHECKING:
            assert isinstance(mesh, Mesh)

        if self.m_IndexBuffer is None and mesh.m_IndexBuffer:
            self.m_IndexBuffer = mesh.m_IndexBuffer

        if self.m_Vertices is None and mesh.m_Vertices:
            self.m_Vertices = vector_list_to_tuples(mesh.m_Vertices)

        if self.m_Normals is None and mesh.m_Normals:
            self.m_Normals = vector_list_to_tuples(mesh.m_Normals)

        if self.m_Tangents is None and mesh.m_Tangents:
            self.m_Tangents = vector_list_to_tuples(mesh.m_Tangents)

        if self.m_UV0 is None and mesh.m_UV:
            self.m_UV0 = vector_list_to_tuples(mesh.m_UV)

        if self.m_UV1 is None and mesh.m_UV1:
            self.m_UV1 = vector_list_to_tuples(mesh.m_UV1)

        if self.m_Colors is None and mesh.m_Colors:
            self.m_Colors = [
                (color.r / 255.0, color.g / 255.0, color.b / 255.0, color.a / 255.0) for color in mesh.m_Colors
            ]

        if self.m_BoneWeights is None and mesh.m_Skin:
            # BoneInfluence == BoneWeight in terms of usage in UnityPy due to int simplification
            self.m_BoneIndices = [
                (skin.boneIndex_0_, skin.boneIndex_1_, skin.boneIndex_2_, skin.boneIndex_3_) for skin in mesh.m_Skin
            ]
            self.m_BoneWeights = [
                (skin.weight_0_, skin.weight_1_, skin.weight_2_, skin.weight_3_) for skin in mesh.m_Skin
            ]

    def copy_from_spriterenderdata(self):
        rd = self.src
        if TYPE_CHECKING:
            assert isinstance(rd, SpriteRenderData)

        if self.m_IndexBuffer is None:
            if rd.m_IndexBuffer:
                self.m_IndexBuffer = rd.m_IndexBuffer
            elif rd.indices:
                self.m_IndexBuffer = rd.indices

        if self.m_Vertices is None and rd.vertices:
            vertices = rd.vertices
            self.m_Vertices = [(v.pos.x, v.pos.y, v.pos.z) for v in vertices]

            if vertices[0].uv is not None:
                self.m_UV0 = [(v.uv.x, v.uv.y) for v in vertices]  # type: ignore

        # if self.m_BindPose is None and rd.m_BindPose:
        #     self.m_BindPose = rd.m_BindPose

    def get_streams(self, m_Channels: list[ChannelInfo], m_VertexCount: int) -> list[StreamInfo]:
        streamCount = 1 + max(x.stream for x in m_Channels)
        m_Streams: list[StreamInfo] = []
        offset = 0
        for s in range(streamCount):
            chnMask = 0
            stride = 0
            for chn, m_Channel in enumerate(m_Channels):
                if m_Channel.stream == s:
                    if m_Channel.dimension > 0:
                        chnMask |= 1 << chn
                        component_size = self.get_channel_component_size(m_Channel)
                        stride += (m_Channel.dimension & 0xF) * component_size

            m_Streams.append(
                StreamInfo(
                    channelMask=chnMask,
                    offset=offset,
                    stride=stride,
                    dividerOp=0,
                    frequency=0,
                )
            )
            offset += m_VertexCount * stride
            offset = (offset + (16 - 1)) & ~(16 - 1)
        return m_Streams

    def get_channels(self, m_Streams: list[StreamInfo]) -> list[ChannelInfo]:
        m_Channels = [
            ChannelInfo(
                dimension=0,
                format=0,
                offset=0,
                stream=0,
            )
            for _ in range(6)
        ]
        for s, m_Stream in enumerate(m_Streams):
            channelMask = bytearray(m_Stream.channelMask)  # BitArray
            offset = 0
            for i in range(6):
                if channelMask[i]:
                    m_Channel = m_Channels[i]
                    m_Channel.stream = s
                    m_Channel.offset = offset
                    if i in [0, 1]:
                        # 0 - kShaderChannelVertex
                        # 1 - kShaderChannelNormal
                        m_Channel.format = 0  # kChannelFormatFloat
                        m_Channel.dimension = 3
                    elif i == 2:  # kShaderChannelColor
                        m_Channel.format = 2  # kChannelFormatColor
                        m_Channel.dimension = 4
                    elif i in [3, 4]:
                        # 3 - kShaderChannelTexCoord0
                        # 4 - kShaderChannelTexCoord1
                        m_Channel.format = 0  # kChannelFormatFloat
                        m_Channel.dimension = 2
                    elif i == 5:  # kShaderChannelTangent
                        m_Channel.format = 0  # kChannelFormatFloat
                        m_Channel.dimension = 4

                    component_size = self.get_channel_component_size(m_Channel)
                    offset += m_Channel.dimension * component_size

        return m_Channels

    def read_vertex_data(self, m_Channels: list[ChannelInfo], m_Streams: list[StreamInfo]) -> None:
        m_VertexData = self.src.m_VertexData
        if m_VertexData is None:
            return

        self.m_VertexCount = m_VertexCount = m_VertexData.m_VertexCount
        # m_VertexDataRaw = m_VertexData.m_DataSize

        for chn, m_Channel in enumerate(m_Channels):
            if m_Channel.dimension == 0:
                continue

            m_Stream = m_Streams[m_Channel.stream]
            # m_StreamData = m_VertexDataRaw[
            #     m_Stream.offset : m_Stream.offset + m_VertexCount * m_Stream.stride
            # ]

            channelMask = bin(m_Stream.channelMask)[::-1]
            if channelMask[chn] == "1":
                if (
                    self.version[0] < 2018 and chn == 2 and m_Channel.format == 2
                ):  # kShaderChannelColor && kChannelFormatColor
                    # new instance to not modify the original
                    m_Channel = ChannelInfo(
                        dimension=4,
                        format=2,
                        offset=m_Channel.offset,
                        stream=m_Channel.stream,
                    )

                component_dtype = self.get_channel_dtype(m_Channel)
                component_byte_size = self.get_channel_component_size(m_Channel)
                # channel_byte_size = m_Channel.dimension * component_byte_size

                swap = self.endianess == "<" and component_byte_size > 1
                channel_dimension = m_Channel.dimension & 0xF

                if UnityPyBoost:
                    componentBytes = UnityPyBoost.unpack_vertexdata(
                        m_VertexData.m_DataSize,
                        component_byte_size,
                        m_VertexCount,
                        m_Stream.offset,
                        m_Stream.stride,
                        m_Channel.offset,
                        channel_dimension,
                        swap,
                    )
                else:
                    componentBytes = bytearray(m_VertexCount * channel_dimension * component_byte_size)

                    vertexBaseOffset = m_Stream.offset + m_Channel.offset
                    for v in range(m_VertexCount):
                        vertexOffset = vertexBaseOffset + m_Stream.stride * v
                        for d in range(channel_dimension):
                            componentOffset = vertexOffset + component_byte_size * d
                            vertexDataSrc = componentOffset
                            componentDataSrc = component_byte_size * (v * channel_dimension + d)
                            buff = m_VertexData.m_DataSize[vertexDataSrc : vertexDataSrc + component_byte_size]
                            if swap:  # swap bytes
                                buff = buff[::-1]
                            componentBytes[componentDataSrc : componentDataSrc + component_byte_size] = buff

                component_data = list(struct.iter_unpack(f">{channel_dimension}{component_dtype}", componentBytes))
                self.assign_channel_vertex_data(chn, component_data)

    def assign_channel_vertex_data(self, channel: int, component_data: list):
        if self.version[0] >= 2018:
            if channel == 0:  # kShaderChannelVertex
                self.m_Vertices = component_data
            elif channel == 1:  # kShaderChannelNormal
                self.m_Normals = component_data
            elif channel == 2:  # kShaderChannelTangent
                self.m_Tangents = component_data
            elif channel == 3:  # kShaderChannelColor
                self.m_Colors = component_data
            elif channel == 4:  # kShaderChannelTexCoord0
                self.m_UV0 = component_data
            elif channel == 5:  # kShaderChannelTexCoord1
                self.m_UV1 = component_data
            elif channel == 6:  # kShaderChannelTexCoord2
                self.m_UV2 = component_data
            elif channel == 7:  # kShaderChannelTexCoord3
                self.m_UV3 = component_data
            elif channel == 8:  # kShaderChannelTexCoord4
                self.m_UV4 = component_data
            elif channel == 9:  # kShaderChannelTexCoord5
                self.m_UV5 = component_data
            elif channel == 10:  # kShaderChannelTexCoord6
                self.m_UV6 = component_data
            elif channel == 11:  # kShaderChannelTexCoord7
                self.m_UV7 = component_data
            # 2018.2 and up
            elif channel == 12:  # kShaderChannelBlendWeight
                self.m_BoneWeights = component_data
            elif channel == 13:  # kShaderChannelBlendIndices
                self.m_BoneIndices = component_data
            else:
                raise ValueError(f"Unknown channel {channel}")
        else:
            if channel == 0:  # kShaderChannelVertex
                self.m_Vertices = component_data
            elif channel == 1:  # kShaderChannelNormal
                self.m_Normals = component_data
            elif channel == 2:  # kShaderChannelColor
                self.m_Colors = component_data
            elif channel == 3:  # kShaderChannelTexCoord0
                self.m_UV0 = component_data
            elif channel == 4:  # kShaderChannelTexCoord1
                self.m_UV1 = component_data
            elif channel == 5:
                if self.version[0] >= 5:  # kShaderChannelTexCoord2
                    self.m_UV2 = component_data
                else:  # kShaderChannelTangent
                    self.m_Tangents = component_data
            elif channel == 6:  # kShaderChannelTexCoord3
                self.m_UV3 = component_data
            elif channel == 7:  # kShaderChannelTangent
                self.m_Tangents = component_data
            else:
                raise ValueError(f"Unknown channel {channel}")

    def get_channel_dtype(self, m_Channel: ChannelInfo):
        if self.version[0] < 2017:
            format = VertexChannelFormat(m_Channel.format)
            component_dtype = VERTEX_CHANNEL_FORMAT_STRUCT_TYPE_MAP[format]
        elif self.version[0] < 2019:
            format = VertexFormat2017(m_Channel.format)
            component_dtype = VERTEX_FORMAT_2017_STRUCT_TYPE_MAP[format]
        else:
            format = VertexFormat(m_Channel.format)
            component_dtype = VERTEX_FORMAT_STRUCT_TYPE_MAP[format]

        return component_dtype

    def get_channel_component_size(self, m_Channel: ChannelInfo):
        dtype = self.get_channel_dtype(m_Channel)
        return struct.Struct(dtype).size

    def decompress_compressed_mesh(self):
        # TODO: m_Triangles????

        version = self.version
        assert isinstance(self.src, Mesh)
        m_CompressedMesh = self.src.m_CompressedMesh

        # Vertex
        self.m_VertexCount = m_VertexCount = m_CompressedMesh.m_Vertices.m_NumItems // 3

        if m_CompressedMesh.m_Vertices.m_NumItems > 0:
            self.m_Vertices = unpack_floats(m_CompressedMesh.m_Vertices, shape=(3,))

        # UV
        if m_CompressedMesh.m_UV.m_NumItems > 0:
            m_UVInfo = m_CompressedMesh.m_UVInfo
            if m_UVInfo is not None and m_UVInfo != 0:
                kInfoBitsPerUV = 4
                kUVDimensionMask = 3
                kUVChannelExists = 4
                kMaxTexCoordShaderChannels = 8

                uvSrcOffset = 0

                for uv_channel in range(kMaxTexCoordShaderChannels):
                    texCoordBits = m_UVInfo >> (uv_channel * kInfoBitsPerUV)
                    texCoordBits &= (1 << kInfoBitsPerUV) - 1
                    if (texCoordBits & kUVChannelExists) != 0:
                        uvDim = 1 + int(texCoordBits & kUVDimensionMask)
                        m_UV = unpack_floats(
                            m_CompressedMesh.m_UV,
                            uvSrcOffset,
                            m_VertexCount * uvDim,
                            shape=(uvDim,),
                        )
                        setattr(self, f"m_UV{uv_channel}", m_UV)
                        uvSrcOffset = uvDim * m_VertexCount
            else:
                self.m_UV0 = unpack_floats(m_CompressedMesh.m_UV, 0, m_VertexCount * 2, shape=(2,))
                if m_CompressedMesh.m_UV.m_NumItems >= m_VertexCount * 4:
                    self.m_UV1 = unpack_floats(
                        m_CompressedMesh.m_UV,
                        m_VertexCount * 2,
                        m_VertexCount * 2,
                        shape=(2,),
                    )

        # BindPose
        if version[0] < 5:  # 5.0 down
            m_BindPoses = m_CompressedMesh.m_BindPoses
            if m_BindPoses and m_BindPoses.m_NumItems > 0:
                self.m_BindPose = unpack_floats(
                    m_BindPoses,
                    shape=(
                        4,
                        4,
                    ),
                )

        # Normal
        if m_CompressedMesh.m_Normals.m_NumItems > 0:
            normalData = unpack_floats(m_CompressedMesh.m_Normals, shape=(2,))
            signs = unpack_ints(m_CompressedMesh.m_NormalSigns)

            normals = zeros(self.m_VertexCount, 3)
            for srcNrm, sign, dstNrm in zip(normalData, signs, normals):
                x, y = srcNrm
                zsqr = 1 - x * x - y * y
                if zsqr >= 0:
                    z = math.sqrt(zsqr)
                    dstNrm[:] = x, y, z
                else:
                    z = 0
                    dstNrm[:] = normalize(x, y, z)
                if sign == 0:
                    dstNrm[2] *= -1
            self.m_Normals = lists_to_tuples(normals)

        # Tangent
        if m_CompressedMesh.m_Tangents.m_NumItems > 0:
            tangentData = unpack_floats(m_CompressedMesh.m_Tangents, shape=(2,))
            signs = unpack_ints(m_CompressedMesh.m_TangentSigns, shape=(2,))

            tangents = zeros(self.m_VertexCount, 4)
            for srcTan, (sign_z, sign_w), dstTan in zip(tangentData, signs, tangents):
                x, y = srcTan
                zsqr = 1 - x * x - y * y
                z = 0
                w = 0
                if zsqr >= 0:
                    z = math.sqrt(zsqr)
                else:
                    x, y, z = normalize(x, y, z)
                if sign_z == 0:
                    z = -z
                w = 1.0 if sign_w > 0 else -1.0
                dstTan[:] = x, y, z, w
            self.m_Tangents = lists_to_tuples(tangents)

        # FloatColor
        if version[0] >= 5:  # 5.0 and up
            m_FloatColors = m_CompressedMesh.m_FloatColors
            if m_FloatColors and m_FloatColors.m_NumItems > 0:
                self.m_Colors = unpack_floats(m_FloatColors, shape=(4,))
        # Skin
        if m_CompressedMesh.m_Weights.m_NumItems > 0:
            weightsData = unpack_ints(m_CompressedMesh.m_Weights)
            boneIndicesData = unpack_ints(m_CompressedMesh.m_BoneIndices)

            vertexIndex = 0
            j = 0
            sum = 0

            boneWeights = zeros(self.m_VertexCount, 4)
            boneIndices = zeros(self.m_VertexCount, 4)

            boneIndicesIterator = iter(boneIndicesData)
            for weight, boneIndex in zip(weightsData, boneIndicesIterator):
                # read bone index and weight
                boneWeights[vertexIndex][j] = weight / 31
                boneIndices[vertexIndex][j] = boneIndex

                j += 1
                sum += weight

                # the weights add up to one, continue with the next vertex.
                if sum >= 31:
                    j = 4
                    # set weights and boneIndices to 0,
                    # already done on init
                    vertexIndex += 1
                    j = 0
                    sum = 0
                # we read three weights, but they don't add up to one. calculate the fourth one, and read
                # missing bone index. continue with next vertex.
                elif j == 3:  #
                    boneWeights[vertexIndex][j] = 1 - sum
                    boneIndices[vertexIndex][j] = next(boneIndicesIterator)

                    vertexIndex += 1
                    j = 0
                    sum = 0

            self.m_BoneWeights = lists_to_tuples(boneWeights)
            self.m_BoneIndices = lists_to_tuples(boneIndices)

        # IndexBuffer
        if m_CompressedMesh.m_Triangles.m_NumItems > 0:  #
            self.m_IndexBuffer = unpack_ints(m_CompressedMesh.m_Triangles)
        # Color
        if m_CompressedMesh.m_Colors and m_CompressedMesh.m_Colors.m_NumItems > 0:
            rgba_colors = unpack_ints(m_CompressedMesh.m_Colors)
            self.m_Colors = [
                (
                    ((rgba >> 24) & 0xFF) / 255,
                    ((rgba >> 16) & 0xFF) / 255,
                    ((rgba >> 8) & 0xFF) / 255,
                    (rgba & 0xFF) / 255,
                )
                for rgba in rgba_colors
            ]

    def get_triangles(self) -> List[List[Tuple[int, ...]]]:
        assert self.m_IndexBuffer is not None
        assert self.src.m_SubMeshes is not None

        submeshes: List[List[Tuple[int, ...]]] = []

        for m_SubMesh in self.src.m_SubMeshes:
            firstIndex = m_SubMesh.firstByte // 2
            if not self.m_Use16BitIndices:
                firstIndex //= 2

            indexCount = m_SubMesh.indexCount
            topology = m_SubMesh.topology

            triangles: List[Tuple[int, ...]]

            if topology == MeshTopology.Triangles:
                triangles = [
                    tuple(self.m_IndexBuffer[i : i + 3]) for i in range(firstIndex, firstIndex + indexCount, 3)
                ]

            elif self.version[0] < 4 or topology == MeshTopology.TriangleStrip:
                triangles = [()] * (indexCount - 2)
                triIndex = 0
                for i in range(firstIndex, firstIndex + indexCount - 2):
                    a, b, c = self.m_IndexBuffer[i : i + 3]
                    # skip degenerates
                    if a == b or a == c or b == c:
                        continue
                    # do the winding flip-flop of strips
                    if (i - firstIndex) & 1:
                        triangles[triIndex] = (b, a, c)
                    else:
                        triangles[triIndex] = (a, b, c)
                    triIndex += 1
                triangles = triangles[:triIndex]
                m_SubMesh.indexCount = len(triangles) * 3

            elif topology == MeshTopology.Quads:
                # one quad is two triangles, so // 4 * 2 = // 2
                triangles = [()] * (indexCount // 2)
                triIndex = 0
                for i in range(firstIndex, firstIndex + indexCount, 4):
                    a, b, c, d = self.m_IndexBuffer[i : i + 4]
                    triangles[triIndex] = (a, b, c)
                    triangles[triIndex + 1] = (a, c, d)
                    triIndex += 2

            else:
                raise ValueError("Failed getting triangles. Submesh topology is lines or points.")

            submeshes.append(triangles)

        return submeshes


# COMPRESSION_BIT_SIZES = {
#     "high": {
#         "vertex": 10,
#         "uv": 8,
#         "normal": 6,
#     },
#     "medium": {
#         "vertex": 16,
#         "uv": 10,
#         "normal": 8
#     },
#     "low": {
#         "vertex": 20,
#         "uv": 16,
#         "normal": 8
#     },
# }



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, Iterable, Union, cast

from PIL import Image, ImageDraw
from PIL.Image import Transform, Transpose

from ..classes import SpriteAtlasData
from ..enums import (
    ClassIDType,
    SpriteMeshType,
    SpritePackingMode,
    SpritePackingRotation,
)
from ..helpers.MeshHelper import MeshHandler
from .Texture2DConverter import get_image_from_texture2d

if TYPE_CHECKING:
    from typing import List, Optional, Tuple

    from ..classes import PPtr, Sprite, Texture2D

try:
    import numpy as np
except ImportError:
    np = None


class SpriteSettings:
    packed: bool
    packingMode: SpritePackingMode
    packingRotation: SpritePackingRotation
    meshType: SpriteMeshType

    def __init__(self, settings_raw):
        self.settingsRaw = settings_raw
        self.packed = bool(self.settingsRaw & 1)  # 1
        self.packingMode = SpritePackingMode((self.settingsRaw >> 1) & 1)  # 1
        self.packingRotation = SpritePackingRotation((self.settingsRaw >> 2) & 0xF)  # 4
        self.meshType = SpriteMeshType((self.settingsRaw >> 6) & 1)  # 1
        # rest of the bits are reserved


def get_image(sprite: Sprite, texture: PPtr[Texture2D], alpha_texture: Optional[PPtr[Texture2D]]) -> Image.Image:
    assert sprite.assets_file, "Sprite assets file is not set!"
    cache = cast(Dict[Any, Any], sprite.assets_file._cache)  # TODO: edit in SerializibleFile
    if alpha_texture:
        cache_id = (texture.path_id, alpha_texture.path_id)
        if cache_id not in cache:
            original_image = get_image_from_texture2d(texture.read(), False)
            alpha_image = get_image_from_texture2d(alpha_texture.read(), False)
            original_image = Image.merge("RGBA", (*original_image.split()[:3], alpha_image.split()[0]))
            cache[cache_id] = original_image
    else:
        cache_id = texture.path_id
        if cache_id not in cache:
            original_image = get_image_from_texture2d(texture.read(), False)
            cache[cache_id] = original_image
    return cache[cache_id]


def get_image_from_sprite(m_Sprite: Sprite) -> Image.Image:
    atlas = None
    if m_Sprite.m_SpriteAtlas:
        atlas = m_Sprite.m_SpriteAtlas.read()
    elif m_Sprite.m_AtlasTags:
        # looks like the direct pointer is empty, let's try to find the Atlas via its name
        assert m_Sprite.assets_file, "Sprite assets file is not set!"
        for obj in m_Sprite.assets_file.objects.values():
            if obj.type == ClassIDType.SpriteAtlas:
                atlas = obj.read()
                if atlas.m_Name == m_Sprite.m_AtlasTags[0]:
                    break
                atlas = None

    if atlas:
        sprite_atlas_data = next(value for key, value in atlas.m_RenderDataMap if key == m_Sprite.m_RenderDataKey)
        assert isinstance(sprite_atlas_data, SpriteAtlasData), "SpriteAtlasData not found!"
    else:
        sprite_atlas_data = m_Sprite.m_RD

    m_Texture2D = sprite_atlas_data.texture
    alpha_texture = sprite_atlas_data.alphaTexture
    texture_rect = sprite_atlas_data.textureRect
    settings_raw = sprite_atlas_data.settingsRaw

    original_image = get_image(m_Sprite, m_Texture2D, alpha_texture)

    sprite_image = original_image.crop(
        (
            texture_rect.x,
            texture_rect.y,
            texture_rect.x + texture_rect.width,
            texture_rect.y + texture_rect.height,
        )
    )

    settings_raw = SpriteSettings(settings_raw)
    if settings_raw.packed == 1:
        rotation = settings_raw.packingRotation
        if rotation == SpritePackingRotation.kSPRFlipHorizontal:
            sprite_image = sprite_image.transpose(Transpose.FLIP_LEFT_RIGHT)
        # spriteImage = RotateFlip(RotateFlipType.RotateNoneFlipX)
        elif rotation == SpritePackingRotation.kSPRFlipVertical:
            sprite_image = sprite_image.transpose(Transpose.FLIP_TOP_BOTTOM)
        # spriteImage.RotateFlip(RotateFlipType.RotateNoneFlipY)
        elif rotation == SpritePackingRotation.kSPRRotate180:
            sprite_image = sprite_image.transpose(Transpose.ROTATE_180)
        # spriteImage.RotateFlip(RotateFlipType.Rotate180FlipNone)
        elif rotation == SpritePackingRotation.kSPRRotate90:
            sprite_image = sprite_image.transpose(Transpose.ROTATE_270)
        # spriteImage.RotateFlip(RotateFlipType.Rotate270FlipNone)

    if settings_raw.packingMode == SpritePackingMode.kSPMTight:
        assert m_Sprite.object_reader, "Sprite object reader is not set!"
        mesh = MeshHandler(m_Sprite.m_RD, m_Sprite.object_reader.version)
        mesh.process()

        if mesh.m_UV0 and any(u or v for u, v in mesh.m_UV0):
            # copy triangles from mesh
            sprite_image = render_sprite_mesh(m_Sprite, mesh, original_image)
        else:
            # create mask to keep only the polygon
            sprite_image = mask_sprite(m_Sprite, mesh, sprite_image)

    return sprite_image.transpose(Transpose.FLIP_TOP_BOTTOM)


def mask_sprite(m_Sprite: Sprite, mesh: MeshHandler, sprite_image: Image.Image) -> Image.Image:
    mask_img = Image.new("1", sprite_image.size, color=0)
    draw = ImageDraw.ImageDraw(mask_img)

    # normalize the points
    #  shift the whole point matrix into the positive space
    #  multiply them with a factor to scale them to the image
    positions = mesh.m_Vertices
    assert positions, "No vertices found in sprite mesh!"
    # find the axis that has only one value - can be removed
    # usually the z axis
    min_x = min(x for x, _y, _z in positions)
    min_y = min(y for _x, y, _z in positions)
    factor = m_Sprite.m_PixelsToUnits
    positions_2d = [((x - min_x) * factor, (y - min_y) * factor) for x, y, _z in positions]

    # generate triangles from the given points
    triangles = [
        (
            positions_2d[a],
            positions_2d[b],
            positions_2d[c],
        )
        for submesh in mesh.get_triangles()
        for a, b, c in submesh
    ]

    for triangle in triangles:
        draw.polygon(triangle, fill=1)

    # apply the mask
    if sprite_image.mode == "RGBA":
        # the image already has an alpha channel,
        # so we have to use composite to keep it
        empty_img = Image.new(sprite_image.mode, sprite_image.size, color=0)
        sprite_image = Image.composite(sprite_image, empty_img, mask_img)
    else:
        # add mask as alpha-channel to keep the polygon clean
        sprite_image.putalpha(mask_img)

    return sprite_image


def render_sprite_mesh(m_Sprite: Sprite, mesh: MeshHandler, texture: Image.Image) -> Image.Image:
    for triangles in mesh.get_triangles():
        positions = mesh.m_Vertices
        if not positions:
            continue
        uv = mesh.m_UV0
        if not uv:
            raise ValueError("No UV coordinates found in sprite mesh!")

        # 2. patch position data
        # 2.1 make positions 2d
        # find the axis that has only one value - can be removed
        # usually the z axis
        axis_values = [[pos[i] for pos in positions] for i in range(3)]
        for i in range(2, -1, -1):
            if len(set(axis_values[i])) == 1:
                break
        else:
            raise ValueError("Can't process 3d sprites!")
        axis_values = axis_values[:i] + axis_values[i + 1 :]
        x_min = min(axis_values[0])
        y_min = min(axis_values[1])
        x_max = max(axis_values[0])
        y_max = max(axis_values[1])

        # 2.2 map positions from middle to top left
        # 2.3 convert relative positions to absolute
        pixels_to_units = m_Sprite.m_PixelsToUnits
        positions_abs = [
            (round((x - x_min) * pixels_to_units), round((y - y_min) * pixels_to_units)) for x, y in zip(*axis_values)
        ]
        width, height = texture.size
        uv_abs = [(round(u * width), round(v * height)) for u, v in uv]

        # 2.4 generate final image size
        size = (
            round((x_max - x_min) * pixels_to_units),
            round((y_max - y_min) * pixels_to_units),
        )
        sprite = Image.new(texture.mode, size)

        for tri in triangles:
            copy_triangle(
                texture,
                tuple(uv_abs[i] for i in tri),  # type: ignore
                sprite,
                tuple(positions_abs[i] for i in tri),  # type: ignore
            )

        return sprite
    else:
        raise ValueError("No triangles found in mesh!")


def copy_triangle(
    src_img: Image.Image,
    src_tri: Tuple[Tuple[int, int], Tuple[int, int], Tuple[int, int]],
    dst_img: Image.Image,
    dst_tri: Tuple[Tuple[int, int], Tuple[int, int], Tuple[int, int]],
) -> None:
    src_off = (
        (src_tri[1][0] - src_tri[0][0], src_tri[1][1] - src_tri[0][1]),
        (src_tri[2][0] - src_tri[0][0], src_tri[2][1] - src_tri[0][1]),
    )
    dst_off = (
        (dst_tri[1][0] - dst_tri[0][0], dst_tri[1][1] - dst_tri[0][1]),
        (dst_tri[2][0] - dst_tri[0][0], dst_tri[2][1] - dst_tri[0][1]),
    )

    # check if transform is necessary by comparing the triangle sizes
    if src_off[0] == dst_off[0] and src_off[1] == dst_off[1]:
        # no transform necessary, just copy the triangle

        # make rectangle that contains the triangle
        # upper_left, _, lower_right = sorted(src_tri)
        upper_left = (
            min(src_tri[0][0], src_tri[1][0], src_tri[2][0]),
            min(src_tri[0][1], src_tri[1][1], src_tri[2][1]),
        )
        lower_right = (
            max(src_tri[0][0], src_tri[1][0], src_tri[2][0]),
            max(src_tri[0][1], src_tri[1][1], src_tri[2][1]),
        )
        src_part = src_img.crop((*upper_left, *lower_right))

        # create mask for triangle
        mask_box = [(x - upper_left[0], y - upper_left[1]) for x, y in src_tri]
        mask = Image.new("1", src_part.size)
        maskdraw = ImageDraw.Draw(mask)
        maskdraw.polygon(mask_box, fill=255)

        # paste triangle into destination image
        dst = (
            int(min(dst_tri[0][0], dst_tri[1][0], dst_tri[2][0])),
            int(min(dst_tri[0][1], dst_tri[1][1], dst_tri[2][1])),
        )
        dst_img.paste(src_part, dst, mask=mask)
    else:
        # transform is necessary, use affine transformation
        # https://stackoverflow.com/a/6959111
        ((x11, x12), (x21, x22), (x31, x32)) = src_tri
        ((y11, y12), (y21, y22), (y31, y32)) = dst_tri

        # Construct matrix M manually
        M = [
            [y11, y12, 1, 0, 0, 0],
            [y21, y22, 1, 0, 0, 0],
            [y31, y32, 1, 0, 0, 0],
            [0, 0, 0, y11, y12, 1],
            [0, 0, 0, y21, y22, 1],
            [0, 0, 0, y31, y32, 1],
        ]

        # Vector y corresponds to the x coordinates in the source triangle
        y = [x11, x21, x31, x12, x22, x32]

        if np:
            A = np.linalg.solve(M, y)
        else:
            # np.lingal.solve - obviously way faster, but numpy will only come with 2.0
            A = linalg_solve(M, y)  # type: ignore

        transformed = src_img.transform(dst_img.size, Transform.AFFINE, A)

        mask = Image.new("1", dst_img.size)
        maskdraw = ImageDraw.Draw(mask)
        maskdraw.polygon(dst_tri, fill=255)

        dst_img.paste(transformed, mask=mask)


def linalg_solve(M: List[List[Union[float, int]]], y: List[Union[float, int]]) -> List[float]:
    # M^-1 * y
    M_i = get_matrix_inverse(M)
    return [sum(M_i[i][j] * y[j] for j in range(len(y))) for i in range(len(M_i))]


def transpose_matrix(m: List[List[float]]) -> Iterable[List[float]]:
    # https://stackoverflow.com/a/39881366
    return map(list, zip(*m))


def get_matrix_minor(m: List[List[float]], i: int, j: int) -> List[List[float]]:
    # https://stackoverflow.com/a/39881366
    return [row[:j] + row[j + 1 :] for row in (m[:i] + m[i + 1 :])]


def get_matrix_determinant(m: List[List[float]]) -> float:
    # https://stackoverflow.com/a/39881366
    # base case for 2x2 matrix
    if len(m) == 2:
        return m[0][0] * m[1][1] - m[0][1] * m[1][0]

    return sum(((-1) ** c) * m[0][c] * get_matrix_determinant(get_matrix_minor(m, 0, c)) for c in range(len(m)))


def get_matrix_inverse(m: List[List[float]]) -> List[List[float]]:
    # https://stackoverflow.com/a/39881366
    determinant = get_matrix_determinant(m)
    # special case for 2x2 matrix:
    if len(m) == 2:
        return [
            [m[1][1] / determinant, -1 * m[0][1] / determinant],
            [-1 * m[1][0] / determinant, m[0][0] / determinant],
        ]

    # find matrix of cofactors
    cofactors = [
        [((-1) ** (r + c)) * get_matrix_determinant(get_matrix_minor(m, r, c)) for c in range(len(m))]
        for r in range(len(m))
    ]
    cofactors = list(transpose_matrix(cofactors))

    return [[c / determinant for c in row] for row in cofactors]


__all__ = ["get_image_from_sprite"]
